package main

import (
	"archive/zip"
	"bytes"
	"context"
	"crypto/hmac"
	"crypto/rand"
	"crypto/sha256"
	"encoding/binary"
	"encoding/base64"
	"encoding/csv"
	"encoding/json"
	"encoding/hex"
	"fmt"
	"hash"
	"html/template"
	"io"
	"log"
	"math"
	"net/http"
	"os"
	"os/exec"
	"path/filepath"
	"regexp"
	"sort"
	"strings"
	"strconv"
	"sync"
	"time"
	"unicode/utf8"
)

type server struct {
	mutex sync.Mutex
	eventsByViewKey map[string]*event
	eventsByEditKey map[string]*event
	eventsBySubmissionsKey map[string]*event
	eventsInDueTimeOrder []*event
	eventsInDueTimeOrderNeedToBeSorted bool
	collectionsByViewKey map[string]*collection
	collectionsByEditKey map[string]*collection
}

type metric struct {
	Name string
	Restrictions string
	Submetrics [3]string
	CustomEvaluatorWASM []byte `json:",omitempty"`

	// for migration; unused if version >= 2.
	PrimaryV1 string `json:"Primary,omitempty"`
	SecondaryV1 string `json:"Secondary,omitempty"`
	TertiaryV1 string `json:"Tertiary,omitempty"`
}

type event struct {
	mutex sync.Mutex
	// protects the submissions file from concurrent access.
	submissionMutex sync.Mutex

	Version int

	ViewKey string
	EditKey string
	SubmissionsKey string

	CreationTime time.Time

	// 'private', 'unlisted', or 'public'.
	Visibility string

	// either '' (default style) or 'post'.
	Style string

	// automated visibility change.
	VisibilityChangeTime time.Time
	VisibilityChange string
	visibilityChangeNotify chan struct{}

	EventTitle string
	PuzzleTitle string
	PuzzleImageURL string
	PuzzleLore template.HTML
	PuzzleFile []byte
	PuzzleFileName string

	Metrics []metric

	// for migration; unused if version >= 3.
	Metric1V2 *metric `json:"Metric1,omitempty"`
	Metric2V2 *metric `json:"Metric2,omitempty"`

	DueTime time.Time

	DisableSimulation bool
	HidePuzzleDownloadLinkAndDisableValidation bool

	Collection *collection
}

type metrics struct {
	DisableSimulation bool
	Metrics []metric
}

func (evt *event) MetricsJSON() ([]byte, error) {
	return json.Marshal(metrics{
		evt.DisableSimulation,
		evt.Metrics,
	})
}

type collection struct {
	Name string
	ViewKey string
	EditKey string

	numberOfEvents int
}

type eventPage struct {
	Event *event
	ValidationMessages []string
	Submissions []previousSubmission
	UniqueSubmitters []string
	SavedSubmitter string
	SavedPronouns string
	SavedDiscord string
	SavedEmail string
	SavedSecretKey string
	ShowingPreviousSubmissions bool
}

type submissionMetadata struct {
	Key string
	TruncatedSolutionName string `json:"SolutionName"`
	SubmissionTime time.Time
	SubmitterName string
	SubmitterPronouns string
	SubmitterDiscord string
	SubmitterEmail string
	SecretKeyHash []byte
	EvaluatedMetrics [][]interface{}
	IsShowcase bool `json:",omitempty"`
	IsDeleted bool `json:",omitempty"`

	// from older submission metadata files.
	EvaluatedMetrics1Legacy *[]interface{} `json:"EvaluatedMetrics1,omitempty"`
	EvaluatedMetrics2Legacy *[]interface{} `json:"EvaluatedMetrics2,omitempty"`

	ByteOffsetInFile int64
	ByteLengthInFile int64
}

type previousSubmission struct {
	Metadata submissionMetadata
	Event *event
	Notes string // only set when editing notes.
	NotesWordCount int
}

func (s submissionMetadata) SolutionFileName(puzzleTitle string) string {
	filename := fmt.Sprintf("%v_%v_%v_%v.solution",
	 strings.ReplaceAll(puzzleTitle, "_", " "), s.SubmissionTime.UnixNano(),
	 strings.ReplaceAll(s.TruncatedSolutionName, "_", " "), s.SubmitterName)
	return filenameSanitize.ReplaceAllLiteralString(filename, "")
}

func (s submissionMetadata) SecretKeyHashTruncated() string {
	if len(s.SecretKeyHash) == 0 {
		return ""
	}
	return hex.EncodeToString(s.SecretKeyHash[:6]) + "\u2026"
}

func (s submissionMetadata) SecretKeyHashFull() string {
	return hex.EncodeToString(s.SecretKeyHash)
}

type submission struct {
	metadata submissionMetadata
	data io.Reader
	error error
}

type frontPageEvent struct {
	ViewKey string
	EventTitle string
	PuzzleImageURL string
	DueTime time.Time
	Collection collection
	Style string
	Icon template.HTML
}

type frontPage struct {
	OngoingEventsByCollection map[collection][]frontPageEvent
	CompletedEventsByCollection map[collection][]frontPageEvent
	Collections []collection
}

type verificationResult struct {
	Results [][]interface{}
	Error string
}

type byDueTime []*event
func (a byDueTime) Len() int { return len(a) }
func (a byDueTime) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a byDueTime) Less(i, j int) bool { return a[i].DueTime.Before(a[j].DueTime) }

var filenameSanitize *regexp.Regexp = regexp.MustCompile(`[<>?:"\|/\x7f]|[^[:ascii:]]|[\x00-\x19]`)
var dataPath string
var staticPath string

func main() {
	workingDirectory, err := os.Getwd()
	if err != nil {
		log.Fatal(err)
	}
	dataPath = filepath.Join(workingDirectory, "data")
	staticPath = filepath.Join(workingDirectory, "static")
	s := &server{
		eventsByViewKey: make(map[string]*event),
		eventsByEditKey: make(map[string]*event),
		eventsBySubmissionsKey: make(map[string]*event),
		collectionsByViewKey: make(map[string]*collection),
		collectionsByEditKey: make(map[string]*collection),
	}
	// load all .json files from the data path as events.
	entries, err := os.ReadDir(dataPath)
	if err != nil {
		log.Fatal(err)
	}
	for _, entry := range entries {
		if filepath.Ext(entry.Name()) != ".json" {
			continue
		}
		func () {
			path := filepath.Join(dataPath, entry.Name())
			file, err := os.Open(path)
			if err != nil {
				log.Print(entry.Name(), err)
				return
			}
			defer file.Close()
			evt := &event{}
			if err = json.NewDecoder(file).Decode(evt); err != nil {
				log.Print(entry.Name(), err)
				return
			}
			if evt.Version == 1 {
				evt.Metric1V2 = &metric{}
				evt.Metric1V2.Restrictions = "default restrictions"
				evt.Metric1V2.Submetrics[0] = evt.Metric1V2.PrimaryV1
				evt.Metric1V2.Submetrics[1] = evt.Metric1V2.SecondaryV1
				evt.Metric1V2.Submetrics[2] = evt.Metric1V2.TertiaryV1
				evt.Metric2V2 = &metric{}
				evt.Metric2V2.Restrictions = "default restrictions"
				evt.Metric2V2.Submetrics[0] = evt.Metric2V2.PrimaryV1
				evt.Metric2V2.Submetrics[1] = evt.Metric2V2.SecondaryV1
				evt.Metric2V2.Submetrics[2] = evt.Metric2V2.TertiaryV1
				evt.Version = 2
			}
			if evt.Version == 2 {
				if evt.Metric1V2 != nil {
					evt.Metrics = append(evt.Metrics, *evt.Metric1V2)
					evt.Metric1V2 = nil
				}
				if evt.Metric2V2 != nil {
					evt.Metrics = append(evt.Metrics, *evt.Metric2V2)
					evt.Metric2V2 = nil
				}
				evt.Version = 3
			}
			s.addEvent(evt)
		}()
	}
	log.Fatal(http.ListenAndServe("127.0.0.1:8001", s))
}

func (s *server) ServeHTTP(response http.ResponseWriter, request *http.Request) {
	components := strings.SplitN(request.URL.Path, "/", 4)
	if len(components) == 4 || len(components) < 2 || components[0] != "" {
		response.WriteHeader(404)
		io.WriteString(response, "404 invalid url")
	} else if len(components) == 2 {
		if components[1] == "" {
			response.Header().Set("Content-Type", "text/html;charset=utf-8")
			if err := templates.ExecuteTemplate(response, "front page", s.frontPage()); err != nil {
				log.Panic(err)
			}
		} else if components[1] == "new" {
			if request.Method != "POST" {
				response.WriteHeader(405)
				io.WriteString(response, "405 please click the button instead of going to /new")
			} else {
				b := make([]byte, 64)
				_, err := rand.Read(b)
				if err != nil {
					response.WriteHeader(500)
					fmt.Fprintf(response, "500 %v", err)
				} else {
					evt := &event{
						Version: 2,
						ViewKey: hex.EncodeToString(b[0:16]),
						EditKey: hex.EncodeToString(b[16:32]),
						SubmissionsKey: hex.EncodeToString(b[32:48]),
						CreationTime: time.Now().UTC(),
						Visibility: "unlisted",
						VisibilityChangeTime: time.Now().UTC(),
						EventTitle: "untitled event",
						PuzzleTitle: "untitled puzzle",
						PuzzleLore: "an <i>awesome</i> puzzle\n",
						PuzzleFileName: "empty.puzzle",
						DueTime: time.Now().UTC().AddDate(0, 0, 7),
						Metrics: []metric{
							metric{ Restrictions: "default restrictions" },
							metric{ Restrictions: "default restrictions" },
						},
					}
					s.addEvent(evt)
					evt.saveToFilesystem()
					response.Header().Add("Location", "/event/" + evt.EditKey)
					response.WriteHeader(303)
				}
			}
		} else if components[1] == "favicon.ico" {
			response.Header().Set("Content-Type", "image/svg+xml")
			io.WriteString(response, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>")
			io.WriteString(response, string(iconForPuzzleData([]byte{}, standaloneListIconAttributes)))
		} else {
			evt := s.getEventIfVisible(s.eventsByViewKey, components[1])
			if evt == nil {
				response.Header().Set("Content-Type", "text/html;charset=utf-8")
				if err := templates.ExecuteTemplate(response, "private", eventPage{}); err != nil {
					log.Panic(err)
				}
				return
			}
			page := eventPage{}
			if request.Method == "POST" {
				// limit total multipart request size to 10M.
				request.Body = http.MaxBytesReader(response, request.Body, 10000000)
				if err := request.ParseMultipartForm(1000); err != nil {
					if err.Error() == "http: request body too large" {
						page.ValidationMessages = append(page.ValidationMessages, "form submission hit size limit (the solution files can't be larger than 10,000,000 bytes)")
					} else {
						page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't parse form data (" + err.Error() + ")")
					}
				} else if solution := request.PostForm.Get("notes-for-solution"); len(solution) != 0 {
					s.postNotesPage(response, request, evt, solution)
					return
				} else if solution := request.PostForm.Get("download-previous-submission"); len(solution) != 0 {
					s.postDownloadPreviousSubmission(response, request, evt, solution)
					return
				} else if solution := request.PostForm.Get("delete-previous-submission"); len(solution) != 0 {
					s.postDeletePreviousSubmission(response, request, evt, solution)
					return
				} else if solution := request.PostForm.Get("really-delete-previous-submission"); len(solution) != 0 {
					s.postReallyDeletePreviousSubmission(response, request, evt, solution, &page)
				} else {
					s.postSubmission(response, request, evt, &page)
				}
			} else {
				retrieveSuccessMessage(response, request, &page)
			}
			s.serveEventPage(response, request, s.eventsByViewKey, components[1], page, "view")
		}
	} else if components[1] == "collection" {
		if components[2] == "" {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		response.Header().Set("Content-Type", "text/html;charset=utf-8")
		events := s.eventsInCollection(components[2])
		if len(events) == 0 {
			// xx it should be easier to unify the collection and empty-collection templates now
			if err := templates.ExecuteTemplate(response, "empty-collection", nil); err != nil {
				log.Panic(err)
			}
		} else {
			if err := templates.ExecuteTemplate(response, "collection", events); err != nil {
				log.Panic(err)
			}
		}
	} else if components[1] == "collection-json" {
		func () {
			s.mutex.Lock()
			defer s.mutex.Unlock()
			c := s.collectionsByEditKey[components[2]]
			if c == nil {
				c = &collection{ EditKey: components[2] }
			}
			response.Header().Set("Content-Type", "application/json")
			json.NewEncoder(response).Encode(c)
		}()
	} else if components[1] == "event" {
		evt := s.getEvent(s.eventsByEditKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		page := eventPage{}
		if request.Method == "POST" {
			s.postSubmissionsTable(response, request, evt, &page)
		} else {
			retrieveSuccessMessage(response, request, &page)
		}
		evt.submissionMutex.Lock()
		defer evt.submissionMutex.Unlock()
		metadata, err := readSubmissionMetadata(evt.EditKey)
		if err != nil {
			response.WriteHeader(500)
			fmt.Fprintf(response, "500 %v", err)
			return
		}
		page.Submissions = make([]previousSubmission, len(metadata))
		page.UniqueSubmitters = make([]string, 0, len(metadata))
		for i, m := range metadata {
			notes := ""
			if m.Key != "" {
				notesBytes, err := os.ReadFile(filepath.Join(dataPath, evt.EditKey + ".notes", m.Key + ".txt"))
				if err == nil {
					notes = string(notesBytes)
				}
			}
			page.Submissions[i] = previousSubmission{
				Metadata: m,
				Event: evt,
				NotesWordCount: len(strings.Fields(notes)),
			}
			if !m.IsDeleted {
				page.UniqueSubmitters = append(page.UniqueSubmitters, m.SubmitterName)
			}
		}
		sort.Slice(page.UniqueSubmitters, func(i, j int) bool { return strings.ToLower(page.UniqueSubmitters[i]) < strings.ToLower(page.UniqueSubmitters[j]) })
		removed := 0
		for i := 1; i < len(page.UniqueSubmitters); i++ {
			if strings.ToLower(page.UniqueSubmitters[i]) == strings.ToLower(page.UniqueSubmitters[i - 1]) {
				removed++
				continue
			}
			page.UniqueSubmitters[i - removed] = page.UniqueSubmitters[i]
		}
		page.UniqueSubmitters = page.UniqueSubmitters[:len(page.UniqueSubmitters) - removed]
		s.serveEventPage(response, request, s.eventsByEditKey, components[2], page, "event")
	} else if components[1] == "edit" {
		page := eventPage{}
		if request.Method == "POST" {
			evt := s.getEvent(s.eventsByEditKey, components[2])
			if evt == nil {
				response.WriteHeader(404)
				io.WriteString(response, "404 invalid url")
				return
			}
			s.postEvent(response, request, evt, &page)
		}
		s.serveEventPage(response, request, s.eventsByEditKey, components[2], page, "edit")
	} else if components[1] == "delete" {
		if request.Method == "POST" {
			evt := s.getEvent(s.eventsByEditKey, components[2])
			if evt == nil {
				response.WriteHeader(404)
				io.WriteString(response, "404 invalid url")
				return
			}
			s.deleteEvent(evt)
			response.Header().Set("Content-Type", "text/html;charset=utf-8")
			if err := templates.ExecuteTemplate(response, "deleted", eventPage{ Event: evt }); err != nil {
				log.Panic(err)
			}
			return
		}
		s.serveEventPage(response, request, s.eventsByEditKey, components[2], eventPage{}, "delete")
	} else if components[1] == "submission" {
		evt := s.getEvent(s.eventsBySubmissionsKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		evt.submissionMutex.Lock()
		defer evt.submissionMutex.Unlock()
		r, err := os.Open(filepath.Join(dataPath, evt.EditKey + ".submissions"))
		if err != nil {
			log.Print(err)
			return
		}
		defer r.Close()
		offset, _ := strconv.ParseInt(request.URL.Query().Get("offset"), 10, 64)
		length, _ := strconv.ParseInt(request.URL.Query().Get("length"), 10, 64)
		response.Header().Set("Content-Type", "application/octet-stream")
		io.Copy(response, io.NewSectionReader(r, offset, length))
	} else if components[1] == "notes" {
		evt := s.getEvent(s.eventsBySubmissionsKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 not found")
			return
		}
		submissionKey := request.URL.Query().Get("submission")
		if submissionKey == "" {
			response.WriteHeader(404)
			io.WriteString(response, "404 not found")
			return
		}
		evt.submissionMutex.Lock()
		defer evt.submissionMutex.Unlock()
		metadata, err := readSubmissionMetadata(evt.EditKey)
		if err != nil {
			response.WriteHeader(500)
			fmt.Fprintf(response, "500 couldn't open metadata")
			return
		}
		for _, m := range metadata {
			if m.IsDeleted {
				continue
			}
			if m.Key == submissionKey {
				http.ServeFile(response, request, filepath.Join(dataPath, evt.EditKey + ".notes", m.Key + ".txt"))
				return
			}
		}
		response.WriteHeader(404)
		io.WriteString(response, "404 not found")
	} else if components[1] == "icon" {
		evt := s.getEventIfVisible(s.eventsByViewKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 not found")
			return
		}
		response.Header().Set("Content-Type", "image/svg+xml")
		evt.mutex.Lock()
		defer evt.mutex.Unlock()
		io.WriteString(response, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>")
		io.WriteString(response, string(iconForPuzzleData(evt.PuzzleFile, standaloneListIconAttributes)))
	} else if components[1] == "csv" {
		evt := s.getEvent(s.eventsBySubmissionsKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		evt.mutex.Lock()
		metrics := evt.Metrics
		evt.mutex.Unlock()
		evt.submissionMutex.Lock()
		defer evt.submissionMutex.Unlock()
		metadata, err := readSubmissionMetadata(evt.EditKey)
		if err != nil {
			return
		}
		response.Header().Set("Content-Type", "text/csv;charset=utf-8")
		response.Header().Set("Cache-Control", "max-age=0, must-revalidate")
		writeCSV(response, metadata, metrics)
	} else if components[1] == "archive" {
		evt := s.getEvent(s.eventsBySubmissionsKey, components[2])
		if evt == nil {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		evt.submissionMutex.Lock()
		defer evt.submissionMutex.Unlock()
		r, err := os.Open(filepath.Join(dataPath, evt.EditKey + ".submissions"))
		if err != nil {
			log.Print(err)
			return
		}
		defer r.Close()
		evt.mutex.Lock()
		puzzleTitle := evt.PuzzleTitle
		metrics := evt.Metrics
		evt.mutex.Unlock()
		serveZipArchive(response, request, evt.EditKey, r, puzzleTitle, metrics)
	} else if components[1] == "static" {
		if components[2] != filepath.Base(components[2]) {
			response.WriteHeader(404)
			io.WriteString(response, "404 invalid url")
			return
		}
		http.ServeFile(response, request, filepath.Join(staticPath, components[2]))
	} else {
		response.WriteHeader(404)
		io.WriteString(response, "404 invalid url")
	}
}

func (s *server) getEvent(m map[string]*event, key string) *event {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	return m[key]
}

func (s *server) getEventIfVisible(m map[string]*event, key string) *event {
	var visibility string
	evt := s.getEvent(m, key)
	if evt != nil {
		evt.mutex.Lock()
		visibility = evt.Visibility
		evt.mutex.Unlock()
	}
	if visibility != "unlisted" && visibility != "public" {
		return nil
	}
	return evt
}

func (s *server) addEvent(evt *event) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	s.eventsByViewKey[evt.ViewKey] = evt
	s.eventsByEditKey[evt.EditKey] = evt
	s.eventsBySubmissionsKey[evt.SubmissionsKey] = evt
	s.eventsInDueTimeOrder = append(s.eventsInDueTimeOrder, evt)
	s.eventsInDueTimeOrderNeedToBeSorted = true
	if evt.Collection != nil {
		s.setCollection(evt, *evt.Collection)
	}
	evt.scheduleVisibilityChange()
}

// requires exclusive access to both s and evt.
func (s *server) setCollection(evt *event, target collection) (validViewKey bool) {
	validViewKey = true
	if evt.Collection != nil && evt.Collection.EditKey != "" {
		evt.Collection.numberOfEvents--
		if evt.Collection.numberOfEvents == 0 {
			if evt.Collection.ViewKey != "" {
				delete(s.collectionsByViewKey, evt.Collection.ViewKey)
			}
			delete(s.collectionsByEditKey, evt.Collection.EditKey)
		}
	}
	if target.EditKey != "" {
		c := s.collectionsByEditKey[target.EditKey]
		if c == nil {
			c = &collection{ EditKey: target.EditKey }
			s.collectionsByEditKey[c.EditKey] = c
		}
		c.Name = target.Name
		if s.collectionsByViewKey[target.ViewKey] != c {
			if s.collectionsByViewKey[target.ViewKey] != nil {
				validViewKey = false
			} else {
				if c.ViewKey != "" {
					delete(s.collectionsByViewKey, c.ViewKey)
				}
				c.ViewKey = target.ViewKey
				if c.ViewKey != "" {
					s.collectionsByViewKey[c.ViewKey] = c
				}
			}
		}
		evt.Collection = c
		c.numberOfEvents++
	} else {
		evt.Collection = nil
	}
	return
}

func (s *server) deleteEvent(evt *event) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	evt.mutex.Lock()
	defer evt.mutex.Unlock()
	log.Printf("removing event %v (%v)", evt.EventTitle, evt.EditKey)
	os.Remove(filepath.Join(dataPath, evt.EditKey + ".submissions"))
	os.Remove(filepath.Join(dataPath, evt.EditKey + ".json"))
	os.Remove(filepath.Join(dataPath, evt.EditKey + ".submissions-json"))
	delete(s.eventsByViewKey, evt.ViewKey)
	delete(s.eventsByEditKey, evt.EditKey)
	delete(s.eventsBySubmissionsKey, evt.SubmissionsKey)
	s.setCollection(evt, collection{})
	n := len(s.eventsInDueTimeOrder)
	// remove the element from eventsInDueTimeOrder.
	// invalidates the sort order.
	for i, v := range s.eventsInDueTimeOrder {
		if v != evt {
			continue
		}
		s.eventsInDueTimeOrder[i] = s.eventsInDueTimeOrder[n-1]
		s.eventsInDueTimeOrder[n-1] = nil
		s.eventsInDueTimeOrder = s.eventsInDueTimeOrder[:n-1]
		s.eventsInDueTimeOrderNeedToBeSorted = true
		break
	}
}

// requires exclusive access to s (via s.mutex).
func (s *server) getEventsInDueTimeOrder() []*event {
	if s.eventsInDueTimeOrderNeedToBeSorted {
		sort.Sort(byDueTime(s.eventsInDueTimeOrder))
		s.eventsInDueTimeOrderNeedToBeSorted = false
	}
	return s.eventsInDueTimeOrder
}

func readSubmissionMetadata(editKey string) ([]submissionMetadata, error) {
	file, err := os.Open(filepath.Join(dataPath, editKey + ".submissions-json"))
	if err != nil {
		// not having a file isn't an error, it just means that nothing has been submitted yet.
		return nil, nil
	}
	defer file.Close()
	var metadata []submissionMetadata
	if err := json.NewDecoder(file).Decode(&metadata); err != nil {
		return nil, err
	}
	for i, _ := range metadata {
		v := &metadata[i]
		if v.EvaluatedMetrics1Legacy != nil {
			v.EvaluatedMetrics = append(v.EvaluatedMetrics, *v.EvaluatedMetrics1Legacy)
		}
		if v.EvaluatedMetrics2Legacy != nil {
			v.EvaluatedMetrics = append(v.EvaluatedMetrics, *v.EvaluatedMetrics2Legacy)
		}
		v.EvaluatedMetrics1Legacy = nil
		v.EvaluatedMetrics2Legacy = nil
	}
	return metadata, nil
}

func writeJSON(filename string, data interface{}) {
	file, err := os.CreateTemp("", "event")
	if err != nil {
		log.Fatal(err)
	}
	if err := json.NewEncoder(file).Encode(data); err != nil {
		log.Fatal(err)
	}
	if err := file.Sync(); err != nil {
		log.Fatal(err)
	}
	if err := file.Close(); err != nil {
		log.Fatal(err)
	}
	if err := os.Rename(file.Name(), filename); err != nil {
		log.Fatal(err)
	}
}

func (e *event) saveToFilesystem() {
	writeJSON(filepath.Join(dataPath, e.EditKey + ".json"), e)
}

func (s *server) serveEventPage(response http.ResponseWriter, request *http.Request, m map[string]*event, key string, page eventPage, templateName string) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	page.Event = m[key]
	if page.Event == nil {
		response.WriteHeader(404)
		io.WriteString(response, "404 invalid url")
		return
	}
	if cookie, _ := request.Cookie("submitter"); cookie != nil {
		page.SavedSubmitter = decodeCookieValue(cookie.Value)
	}
	if cookie, _ := request.Cookie("pronouns"); cookie != nil {
		page.SavedPronouns = decodeCookieValue(cookie.Value)
	}
	if cookie, _ := request.Cookie("discord"); cookie != nil {
		page.SavedDiscord = decodeCookieValue(cookie.Value)
	}
	if cookie, _ := request.Cookie("email"); cookie != nil {
		page.SavedEmail = decodeCookieValue(cookie.Value)
	}
	if cookie, _ := request.Cookie("secret-key"); cookie != nil {
		page.SavedSecretKey = decodeCookieValue(cookie.Value)
	}
	response.Header().Set("Content-Type", "text/html;charset=utf-8")
	// ensure the lore html can't run scripts.
	response.Header().Set("Content-Security-Policy", "script-src 'self'")
	if err := templates.ExecuteTemplate(response, templateName, page); err != nil {
		log.Panic(err)
	}
}

func (s *server) frontPage() (page frontPage) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	page.OngoingEventsByCollection = map[collection][]frontPageEvent{}
	page.CompletedEventsByCollection = map[collection][]frontPageEvent{}
	for _, evt := range s.getEventsInDueTimeOrder() {
		func (){
			evt.mutex.Lock()
			defer evt.mutex.Unlock()
			if evt.Visibility != "public" {
				return
			}
			var key collection
			if evt.Collection != nil && evt.Collection.Name != "" {
				key = *evt.Collection
				if len(page.OngoingEventsByCollection[key]) == 0 && len(page.CompletedEventsByCollection[key]) == 0 {
					page.Collections = append(page.Collections, key)
				}
			}
			e := frontPageEvent{
				ViewKey: evt.ViewKey,
				EventTitle: evt.EventTitle,
				PuzzleImageURL: evt.PuzzleImageURL,
				DueTime: evt.DueTime,
				Collection: key,
				Style: evt.Style,
				Icon: iconForPuzzleData(evt.PuzzleFile, frontPageListIconAttributes),
			}
			if time.Since(evt.DueTime) < 0 {
				page.OngoingEventsByCollection[key] = append(page.OngoingEventsByCollection[key], e)
			} else {
				page.CompletedEventsByCollection[key] = append(page.CompletedEventsByCollection[key], e)
			}
		}()
	}
	n := len(page.Collections) - 1
	for i := 0; i < n - i; i++ {
		page.Collections[i], page.Collections[n - i] = page.Collections[n - i], page.Collections[i]
	}
	return
}

func (s *server) eventsInCollection(collectionViewKey string) (events []frontPageEvent) {
	s.mutex.Lock()
	defer s.mutex.Unlock()
	for _, evt := range s.getEventsInDueTimeOrder() {
		func (){
			evt.mutex.Lock()
			defer evt.mutex.Unlock()
			if evt.Visibility != "public" {
				return
			} else if evt.Collection == nil || evt.Collection.ViewKey != collectionViewKey {
				return
			}
			events = append(events, frontPageEvent{
				ViewKey: evt.ViewKey,
				EventTitle: evt.EventTitle,
				PuzzleImageURL: evt.PuzzleImageURL,
				DueTime: evt.DueTime,
				Collection: *evt.Collection,
				Style: evt.Style,
				Icon: iconForPuzzleData(evt.PuzzleFile, frontPageListIconAttributes),
			})
		}()
	}
	return
}

func writeCSV(w io.Writer, metadata []submissionMetadata, metrics []metric) error {
	cw := csv.NewWriter(w)
	for _, m := range metadata {
		if m.IsDeleted {
			continue
		}
		row := []string{
			m.SubmitterName,
			m.SubmitterPronouns,
			m.SubmitterDiscord,
			m.SecretKeyHashFull(),
			m.TruncatedSolutionName,
		}
		for i, v := range metrics {
			evaluated := []interface{}{}
			if i < len(m.EvaluatedMetrics) {
				evaluated = m.EvaluatedMetrics[i]
			}
			for j := 0; j < len(v.Submetrics); j++ {
				if j < len(evaluated) {
					row = append(row, fmtResult(evaluated[j]))
				} else {
					row = append(row, "")
				}
			}
		}
		row = append(row, m.SubmissionTime.Format("January 2, 2006 at 3:04 PM"))
		row = append(row, fmt.Sprintf("%v", m.SubmissionTime.UnixNano()))
		if m.IsShowcase {
			row = append(row, "yes")
		} else {
			row = append(row, "")
		}
		if err := cw.Write(row); err != nil {
			return err
		}
	}
	cw.Flush()
	return nil
}

func serveZipArchive(response http.ResponseWriter, request *http.Request, editKey string, r *os.File, puzzleTitle string, metrics []metric) error {
	response.Header().Set("Content-Type", "application/zip")
	zw := zip.NewWriter(response)
	w, err := zw.Create("data.csv")
	if err != nil {
		return err
	}
	metadata, err := readSubmissionMetadata(editKey)
	if err != nil {
		return err
	}
	writeCSV(w, metadata, metrics)
	for _, m := range metadata {
		if m.IsDeleted {
			continue
		}
		filename := m.SolutionFileName(puzzleTitle)
		if m.IsShowcase {
			filename = "showcase/" + filename
		} else {
			filename = "submissions/" + filename
		}
		f, err := zw.Create(filename)
		if err != nil {
			return err
		}
		if _, err := io.Copy(f, io.NewSectionReader(r, m.ByteOffsetInFile, m.ByteLengthInFile)); err != nil {
			return err
		}
		notes, err := os.Open(filepath.Join(dataPath, editKey + ".notes", m.Key + ".txt"))
		if err == nil {
			f, err = zw.Create(filename + ".notes.txt")
			if err != nil {
				notes.Close()
				return err
			}
			if _, err := io.Copy(f, notes); err != nil {
				notes.Close()
				return err
			}
			notes.Close()
		}
	}
	zw.Close()
	return nil
}

func getFormValidatingLength(request *http.Request, key string, name string, defaultValue string, maxRuneCount int, validationMessages []string) (string, []string) {
	v := request.PostForm.Get(key)
	if utf8.RuneCountInString(v) > maxRuneCount {
		return defaultValue, append(validationMessages, fmt.Sprintf("%s can be at most %d characters long, but it was %d", name, maxRuneCount, utf8.RuneCountInString(v)))
	}
	return v, validationMessages
}

func redirectWithSuccessMessage(response http.ResponseWriter, page *eventPage, path string, successMessage string) {
	// successful submissions are redirected to allow reloading the page without
	// a "reloading will submit the form again" warning.  pass the success
	// message along using a cookie.
	if page != nil {
		page.ValidationMessages = append(page.ValidationMessages, successMessage)
	}
	cookie := &http.Cookie{ Name: "success-message", Value: encodeCookieValue(successMessage), MaxAge: 30, Path: "/" }
	http.SetCookie(response, cookie)
	response.Header().Add("Location", path)
	response.WriteHeader(303)
}

func retrieveSuccessMessage(response http.ResponseWriter, request *http.Request, page *eventPage) {
	if cookie, _ := request.Cookie("success-message"); cookie != nil {
		// clear the cookie now that we've displayed the message.
		http.SetCookie(response, &http.Cookie{ Name: "success-message", MaxAge: -1 })
		page.ValidationMessages = append(page.ValidationMessages, decodeCookieValue(cookie.Value))
	}
}

func parseLength(data []byte) (int, []byte) {
	n := 0
	shift := 0
	for len(data) > 0 {
		a := data[0]
		data = data[1:]
		n |= (int(a) & 0x7f) << shift
		shift += 7
		if a & 0x80 == 0 {
			break
		}
	}
	return n, data
}

func puzzleName(puzzleData []byte) string {
	if len(puzzleData) < 5 {
		return ""
	}
	if puzzleData[0] != 3 {
		return ""
	}
	n, puzzleData := parseLength(puzzleData[4:])
	if len(puzzleData) < n {
		return ""
	}
	return string(puzzleData[:n])
}

func solutionPuzzleFileName(solutionData []byte) string {
	if len(solutionData) < 5 {
		return ""
	}
	if solutionData[0] != 7 {
		return ""
	}
	n, solutionData := parseLength(solutionData[4:])
	if len(solutionData) < n {
		return ""
	}
	return string(solutionData[:n])
}

func truncatedSolutionName(solutionData []byte) string {
	if len(solutionData) < 5 {
		return ""
	}
	if solutionData[0] != 7 {
		return ""
	}
	n, solutionData := parseLength(solutionData[4:])
	if len(solutionData) < n {
		return ""
	}
	m, solutionData := parseLength(solutionData[n:])
	if len(solutionData) < m {
		return ""
	}
	if m > 83 {
		return fmt.Sprintf("%s\u2026", solutionData[:80])
	} else {
		return string(solutionData[:m])
	}
}

// must be called with evt.mutex held.
func (evt *event) scheduleVisibilityChange() {
	if evt.visibilityChangeNotify != nil {
		evt.visibilityChangeNotify <- struct{}{}
	} else if evt.VisibilityChange != "" {
		evt.visibilityChangeNotify = make(chan struct{})
		go func () {
			var timer *time.Timer
			for {
				var notify chan struct{}
				var d time.Duration
				func () {
					evt.mutex.Lock()
					defer evt.mutex.Unlock()
					notify = evt.visibilityChangeNotify
					if evt.VisibilityChange == "" {
						d = 0
					} else {
						d = time.Until(evt.VisibilityChangeTime)
						if d <= 0 {
							evt.Visibility = evt.VisibilityChange
							evt.VisibilityChange = ""
							evt.visibilityChangeNotify = nil
							evt.saveToFilesystem()
						}
					}
				}()
				if d <= 0 {
					return
				} else if timer == nil {
					timer = time.NewTimer(d)
				} else {
					timer.Reset(d)
				}
				select {
				case <-notify:
					if !timer.Stop() {
						<-timer.C
					}
				case <-timer.C:
					timer.Stop()
				}
			}
		}()
	}
}

func writeNote(editKey string, solutionKey string, notes string) error {
	notesDirectory := filepath.Join(dataPath, editKey + ".notes")
	notesPath := filepath.Join(notesDirectory, solutionKey + ".txt")
	if err := os.MkdirAll(notesDirectory, 0755); err != nil {
		return err
	}
	if err := os.WriteFile(notesPath, []byte(notes), 0644); err != nil {
		return err
	}
	if notes == "" {
		if err := os.Remove(notesPath); err != nil {
			return err
		}
	}
	return nil
}

func (s *server) postNotesPage(response http.ResponseWriter, request *http.Request, evt *event, solutionKey string) {
	if solutionKey == "new submission" {
		notes := request.PostForm.Get("existing-notes")
		p := previousSubmission{
			Event: evt,
			Notes: notes,
			NotesWordCount: len(strings.Fields(notes)),
		}
		if err := json.Unmarshal([]byte(request.PostForm.Get("metadata")), &p.Metadata); err != nil {
			response.WriteHeader(500)
			io.WriteString(response, "500 internal error: couldn't decode submission metadata form data")
			return
		}
		evt.mutex.Lock()
		defer evt.mutex.Unlock()
		response.Header().Set("Content-Type", "text/html;charset=utf-8")
		if err := templates.ExecuteTemplate(response, "edit notes", p); err != nil {
			log.Panic(err)
		}
		return
	}
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		response.WriteHeader(500)
		io.WriteString(response, "500 internal error: couldn't decode submission metadata file")
		return
	}
	for _, m := range metadata {
		if m.IsDeleted {
			continue
		}
		if solutionKey == m.Key {
			notesDirectory := filepath.Join(dataPath, evt.EditKey + ".notes")
			notesPath := filepath.Join(notesDirectory, m.Key + ".txt")
			if len(request.PostForm["notes"]) != 0 {
				if err := writeNote(evt.EditKey, m.Key, request.PostForm.Get("notes")); err != nil {
					response.WriteHeader(500)
					io.WriteString(response, "500 internal error: couldn't write notes file")
					return
				}
				redirectWithSuccessMessage(response, nil, "/" + evt.ViewKey, "notes saved for " + m.TruncatedSolutionName)
				return
			}
			notes := ""
			notesBytes, err := os.ReadFile(notesPath)
			if err == nil {
				notes = string(notesBytes)
			}
			p := previousSubmission{
				Metadata: m,
				Event: evt,
				Notes: notes,
				NotesWordCount: len(strings.Fields(notes)),
			}
			evt.mutex.Lock()
			defer evt.mutex.Unlock()
			response.Header().Set("Content-Type", "text/html;charset=utf-8")
			if err := templates.ExecuteTemplate(response, "edit notes", p); err != nil {
				log.Panic(err)
			}
			return
		}
	}
	response.WriteHeader(404)
	io.WriteString(response, "404 solution not found")
}

func (s *server) postDownloadPreviousSubmission(response http.ResponseWriter, request *http.Request, evt *event, solutionKey string) {
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		response.WriteHeader(500)
		io.WriteString(response, "500 internal error: couldn't decode submission metadata file")
		return
	}
	for _, m := range metadata {
		if m.IsDeleted {
			continue
		}
		if solutionKey == m.Key {
			evt.submissionMutex.Lock()
			defer evt.submissionMutex.Unlock()
			r, err := os.Open(filepath.Join(dataPath, evt.EditKey + ".submissions"))
			if err != nil {
				response.WriteHeader(500)
				io.WriteString(response, "500 internal error: couldn't open submissions file")
				return
			}
			defer r.Close()
			response.Header().Set("Content-Disposition", fmt.Sprintf(`inline; filename="%s"`, m.SolutionFileName(evt.PuzzleTitle)))
			http.ServeContent(response, request, m.SolutionFileName(evt.PuzzleTitle), time.Time{}, io.NewSectionReader(r, m.ByteOffsetInFile, m.ByteLengthInFile))
			return
		}
	}
	response.WriteHeader(404)
	io.WriteString(response, "404 solution not found")
}

func (s *server) postDeletePreviousSubmission(response http.ResponseWriter, request *http.Request, evt *event, solutionKey string) {
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		response.WriteHeader(500)
		io.WriteString(response, "500 internal error: couldn't decode submission metadata file")
		return
	}
	for _, m := range metadata {
		if m.IsDeleted {
			continue
		}
		if solutionKey == m.Key {
			submission := previousSubmission{
				Metadata: m,
				Event: evt,
			}
			response.Header().Set("Content-Type", "text/html;charset=utf-8")
			if err := templates.ExecuteTemplate(response, "delete submission", submission); err != nil {
				log.Panic(err)
			}
			return
		}
	}
	response.WriteHeader(404)
	io.WriteString(response, "404 solution not found")
}

func (s *server) postReallyDeletePreviousSubmission(response http.ResponseWriter, request *http.Request, evt *event, solutionKey string, page *eventPage) {
	evt.submissionMutex.Lock()
	defer evt.submissionMutex.Unlock()
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		response.WriteHeader(500)
		io.WriteString(response, "500 internal error: couldn't decode submission metadata file")
		return
	}
	for i, m := range metadata {
		if m.IsDeleted {
			continue
		}
		if solutionKey == m.Key {
			metadata[i].IsDeleted = true
			writeJSON(filepath.Join(dataPath, evt.EditKey + ".submissions-json"), metadata)
			redirectWithSuccessMessage(response, page, request.URL.Path, "submission deleted.")
			return
		}
	}
	response.WriteHeader(404)
	io.WriteString(response, "404 solution not found")
}

func decodeCookieValue(value string) string {
	if strings.HasPrefix(value, "!base64!") {
		decoded, _ := base64.URLEncoding.DecodeString(value[len("!base64!"):])
		return string(decoded)
	} else {
		return value
	}
}

func encodeCookieValue(value string) string {
	return "!base64!" + base64.URLEncoding.EncodeToString([]byte(value))
}

func (s *server) postSubmission(response http.ResponseWriter, request *http.Request, evt *event, page *eventPage) {
	saveSubmitterInfo := len(request.PostForm.Get("save")) > 0
	if !saveSubmitterInfo {
		// clear the cookies.
		http.SetCookie(response, &http.Cookie{ Name: "submitter", MaxAge: -1 })
		http.SetCookie(response, &http.Cookie{ Name: "pronouns", MaxAge: -1 })
		http.SetCookie(response, &http.Cookie{ Name: "discord", MaxAge: -1 })
		http.SetCookie(response, &http.Cookie{ Name: "email", MaxAge: -1 })
		http.SetCookie(response, &http.Cookie{ Name: "secret-key", MaxAge: -1 })
		request.Header.Del("Cookie")
	}
	submission := submissionMetadata{
		SubmitterName: request.PostForm.Get("submitter"),
		SubmitterPronouns: request.PostForm.Get("pronouns"),
		SubmitterDiscord: request.PostForm.Get("discord"),
		SubmitterEmail: request.PostForm.Get("email"),
		IsShowcase: len(request.PostForm.Get("showcase")) > 0,
	}
	secretKey := request.PostForm.Get("secret-key")
	if len(secretKey) > 0 {
		if len(secretKey) < 8 {
			page.ValidationMessages = append(page.ValidationMessages, "secret key must be at least 8 characters")
			return
		}
		if len(secretKey) > 300 {
			page.ValidationMessages = append(page.ValidationMessages, "secret key must be less than 300 characters long")
			return
		}
		salt, err := hex.DecodeString(evt.SubmissionsKey)
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: this event's submissions key isn't a valid hexadecimal number")
			return
		}
		salt = append(salt, []byte(submission.SubmitterName)...)
		submission.SecretKeyHash = pbkdf2([]byte(secretKey), salt, 4096, 20, sha256.New)
	}
	if len(submission.SubmitterName) == 0 {
		page.ValidationMessages = append(page.ValidationMessages, "submissions must have an associated submitter name")
		return
	}
	if utf8.RuneCountInString(submission.SubmitterName) > 48 {
		page.ValidationMessages = append(page.ValidationMessages, "submitter name must be less than 48 characters long")
		return
	}
	if utf8.RuneCountInString(submission.SubmitterDiscord) > 48 {
		page.ValidationMessages = append(page.ValidationMessages, "submitter discord tag must be less than 48 characters long")
		return
	}
	if utf8.RuneCountInString(submission.SubmitterEmail) > 48 {
		page.ValidationMessages = append(page.ValidationMessages, "submitter email address must be less than 48 characters long")
		return
	}
	if saveSubmitterInfo {
		// save these form fields as cookies.
		addCookie := func (name string) {
			c := &http.Cookie{ Name: name, Value: encodeCookieValue(request.PostForm.Get(name)), MaxAge: 31536000 }
			http.SetCookie(response, c)
			// also add the cookie to the request so that serveEditPage() will see it.
			request.AddCookie(c)
		}
		request.Header.Del("Cookie")
		addCookie("submitter")
		addCookie("pronouns")
		addCookie("discord")
		addCookie("email")
		addCookie("secret-key")
	}
	evt.mutex.Lock()
	submittable := evt.Style != "post"
	puzzle := evt.PuzzleFile
	puzzleFileName := evt.PuzzleFileName
	validatePuzzleFileName := !evt.HidePuzzleDownloadLinkAndDisableValidation
	dueTime := evt.DueTime
	metrics, err := evt.MetricsJSON()
	evt.mutex.Unlock()
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to encode metrics")
		return
	}
	if !submittable {
		page.ValidationMessages = append(page.ValidationMessages, "event does not accept submissions")
		return
	}
	if len(puzzle) == 0 {
		page.ValidationMessages = append(page.ValidationMessages, "event doesn't have a puzzle file set")
		return
	}
	if len(request.PostForm.Get("show-previous")) > 0 {
		page.ShowingPreviousSubmissions = true
		if len(submission.SecretKeyHash) == 0 {
			return
		}
		// we're showing previous submissions, not making a new submission.
		metadata, err := readSubmissionMetadata(evt.EditKey)
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't decode submission metadata file (" + err.Error() + ")")
			return
		}
		page.Submissions = make([]previousSubmission, len(metadata))
		numberOfSubmissions := 0
		for _, m := range metadata {
			if m.IsDeleted {
				continue
			}
			if bytes.Equal(submission.SecretKeyHash, m.SecretKeyHash) {
				numberOfSubmissions++
				notes := ""
				notesBytes, err := os.ReadFile(filepath.Join(dataPath, evt.EditKey + ".notes", m.Key + ".txt"))
				if err == nil {
					notes = string(notesBytes)
				}
				page.Submissions[len(metadata) - numberOfSubmissions] = previousSubmission{
					Metadata: m,
					Event: evt,
					NotesWordCount: len(strings.Fields(notes)),
				}
			}
		}
		page.Submissions = page.Submissions[len(metadata) - numberOfSubmissions:]
		return
	}
	files := request.MultipartForm.File["solution"]
	if len(files) == 0 {
		page.ValidationMessages = append(page.ValidationMessages, "no solution files were attached")
		return
	}
	evt.submissionMutex.Lock()
	defer evt.submissionMutex.Unlock()
	archive, err := os.OpenFile(filepath.Join(dataPath, evt.EditKey + ".submissions"), os.O_CREATE|os.O_WRONLY, 0644)
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't open submissions file (" + err.Error() + ")")
		return
	}
	offset, err := archive.Seek(0, io.SeekEnd)
	if err != nil {
		archive.Close()
		page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't seek submissions file (" + err.Error() + ")")
		return
	}
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't decode submission metadata file (" + err.Error() + ")")
		return
	}
	defer func () {
		// if there was an error, revert the file to its old length and close it.
		if archive == nil {
			return
		}
		page.ValidationMessages = append(page.ValidationMessages, "due to the above issues, no solution files were submitted")
		archive.Truncate(offset)
		archive.Close()
	}()
	numberOfSubmissions := 0
	ctx, cancel := context.WithTimeout(context.Background(), 2 * time.Minute)
	defer cancel()
	cmd := exec.CommandContext(ctx, "node", filepath.Join(staticPath, "verify.js"))
	cmdIn, err := cmd.StdinPipe()
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open pipe (" + err.Error() + ")")
		return
	}
	cmdOut, err := cmd.StdoutPipe()
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open pipe (" + err.Error() + ")")
		return
	}
	resultDecoder := json.NewDecoder(cmdOut)
	cmdErr, err := cmd.StderrPipe()
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open pipe (" + err.Error() + ")")
		return
	}
	if err := cmd.Start(); err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: cannot start validator process (" + err.Error() + ")")
		return
	}
	defer func () {
		if cmdIn != nil {
			cmdIn.Close()
			cmdIn = nil
		}
		cmd.Wait()
	}()
	encodeDataWithLength := func (data []byte) error {
		if err := binary.Write(cmdIn, binary.LittleEndian, uint32(len(data))); err != nil {
			return err
		}
		if _, err := cmdIn.Write(data); err != nil {
			return err
		}
		return nil
	}
	if err := encodeDataWithLength(puzzle); err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to write to pipe (" + err.Error() + ")")
		return
	}
	if err := encodeDataWithLength(metrics); err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to write to pipe (" + err.Error() + ")")
		return
	}
	for _, file := range files {
		if file.Size > 1000000 {
			page.ValidationMessages = append(page.ValidationMessages, "solution file '" + file.Filename + "' can't be larger than 1,000,000 bytes")
			return
		}
		f, err := file.Open()
		defer f.Close()
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open uploaded file (" + err.Error() + ")")
			return
		}
		b := make([]byte, 16)
		if _, err := rand.Read(b); err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to generate key (" + err.Error() + ")")
			return
		}
		submission.Key = hex.EncodeToString(b)
		submission.SubmissionTime = time.Now().UTC()
		if submission.SubmissionTime.After(dueTime) && !submission.IsShowcase {
			page.ValidationMessages = append(page.ValidationMessages, "the deadline for this event has passed")
			return
		}
		data, err := io.ReadAll(f)
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to read uploaded file")
			return
		}
		submission.TruncatedSolutionName = truncatedSolutionName(data)
		if len(submission.TruncatedSolutionName) == 0 {
			page.ValidationMessages = append(page.ValidationMessages, "file '" + file.Filename + "' doesn't seem to be an opus magnum solution file")
			return
		}
		if validatePuzzleFileName && solutionPuzzleFileName(data) + ".puzzle" != puzzleFileName {
			page.ValidationMessages = append(page.ValidationMessages, "file '" + file.Filename + "' is a solution to a different puzzle, '" + solutionPuzzleFileName(data) + ".puzzle'")
			return
		}
		if !submission.IsShowcase {
			if err := encodeDataWithLength(data); err != nil {
				page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to write to pipe (" + err.Error() + ")")
				return
			}
			var result verificationResult
			if err := resultDecoder.Decode(&result); err != nil {
				page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to read from pipe (" + err.Error() + ")")
				return
			}
			if len(result.Error) > 0 {
				page.ValidationMessages = append(page.ValidationMessages, "error while verifying '" + file.Filename + "': " + result.Error)
				return
			}
			submission.EvaluatedMetrics = result.Results
		}
		submission.ByteLengthInFile = int64(len(data))
		submission.ByteOffsetInFile, err = archive.Seek(0, io.SeekCurrent)
		if err != nil {
			archive.Close()
			page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't seek submissions file (" + err.Error() + ")")
			return
		}
		if _, err = archive.Write(data); err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to write to submissions file")
			return
		}
		metadata = append(metadata, submission)
		if notes := request.PostForm.Get("notes-content-" + strconv.Itoa(numberOfSubmissions)); notes != "" {
			if err := writeNote(evt.EditKey, submission.Key, notes); err != nil {
				page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to write notes file")
				return
			}
		}
		numberOfSubmissions += 1
	}
	cmdIn.Close()
	cmdIn = nil
	io.Copy(os.Stderr, cmdErr)
	archive.Close()
	archive = nil
	writeJSON(filepath.Join(dataPath, evt.EditKey + ".submissions-json"), metadata)
	if numberOfSubmissions == 1 {
		redirectWithSuccessMessage(response, page, request.URL.Path, "your solution was submitted successfully!")
	} else {
		redirectWithSuccessMessage(response, page, request.URL.Path, fmt.Sprintf("your %d solutions were submitted successfully!", numberOfSubmissions))
	}
}

func (s *server) postSubmissionsTable(response http.ResponseWriter, request *http.Request, evt *event, page *eventPage) {
	if err := request.ParseForm(); err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't parse form data (" + err.Error() + ")")
		return
	}
	evt.submissionMutex.Lock()
	defer evt.submissionMutex.Unlock()
	metadata, err := readSubmissionMetadata(evt.EditKey)
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't open submission metadata file (" + err.Error() + ")")
		return
	}
	for i := 0; i < len(metadata); i++ {
		hasData := false
		update := func (key string, dest *string) {
			data := request.PostForm[key]
			if len(data) > 0 {
				if utf8.RuneCountInString(data[0]) > 48 {
					page.ValidationMessages = append(page.ValidationMessages, "field data was more than 48 characters long; ignoring")
				} else {
					*dest = data[0]
				}
				hasData = true
			}
		}
		update(fmt.Sprintf("%d.name", i), &metadata[i].SubmitterName)
		update(fmt.Sprintf("%d.pronouns", i), &metadata[i].SubmitterPronouns)
		update(fmt.Sprintf("%d.discord", i), &metadata[i].SubmitterDiscord)
		update(fmt.Sprintf("%d.email", i), &metadata[i].SubmitterEmail)
		if !hasData {
			continue
		}
		metadata[i].IsShowcase = len(request.PostForm.Get(fmt.Sprintf("%d.showcase", i))) > 0;
		metadata[i].IsDeleted = len(request.PostForm.Get(fmt.Sprintf("%d.deleted", i))) > 0;
	}
	writeJSON(filepath.Join(dataPath, evt.EditKey + ".submissions-json"), metadata)
	redirectWithSuccessMessage(response, page, request.URL.Path, "submission data updated successfully!")
}

func (s *server) postEvent(response http.ResponseWriter, request *http.Request, evt *event, page *eventPage) {
	// limit total multipart request size to 100k.
	request.Body = http.MaxBytesReader(response, request.Body, 100000)
	if err := request.ParseMultipartForm(100000); err != nil {
		if err.Error() == "http: request body too large" {
			page.ValidationMessages = append(page.ValidationMessages, "form submission hit size limit (the puzzle file can't be larger than 100,000 bytes)")
		} else {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: couldn't parse form data (" + err.Error() + ")")
		}
		return
	}
	evt.mutex.Lock()
	defer evt.mutex.Unlock()
	now := time.Now()
	var viewKey string
	viewKey, page.ValidationMessages = getFormValidatingLength(request, "view-key", "event id", evt.ViewKey, 64, page.ValidationMessages)
	if viewKey != evt.ViewKey {
		if len(viewKey) == 0 {
			page.ValidationMessages = append(page.ValidationMessages, "event id can't be empty")
		} else if strings.Contains(viewKey, "/") {
			page.ValidationMessages = append(page.ValidationMessages, "event id can't include slashes")
		} else {
			func () {
				s.mutex.Lock()
				defer s.mutex.Unlock()
				if s.eventsByViewKey[viewKey] != nil {
					// xx remove conflicting events that have been completed?
					page.ValidationMessages = append(page.ValidationMessages, "event id is already in use by another event")
				} else {
					s.eventsByViewKey[evt.ViewKey] = nil
					evt.ViewKey = viewKey
					s.eventsByViewKey[viewKey] = evt
				}
			}()
		}
	}
	evt.Visibility = request.PostForm.Get("visibility")
	if evt.Visibility != "private" && evt.Visibility != "unlisted" && evt.Visibility != "public" {
		page.ValidationMessages = append(page.ValidationMessages, "invalid visibility '" + evt.Visibility + "': setting to private")
		evt.Visibility = "private"
	}
	visibilityChange := request.PostForm.Get("visibility-change")
	changeTime, err := time.Parse("2006-01-02@15:04", fmt.Sprintf("%s@%s", request.PostForm.Get("visibility-change-date"), request.PostForm.Get("visibility-change-time")))
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "invalid event visibility automation date")
	} else if visibilityChange != "" && visibilityChange != "private" && visibilityChange != "unlisted" && visibilityChange != "public" {
		page.ValidationMessages = append(page.ValidationMessages, "invalid visibility '" + visibilityChange + "' in automation")
	} else if visibilityChange == evt.VisibilityChange && changeTime == evt.VisibilityChangeTime {
		// no change.
	} else if visibilityChange != "" && changeTime.Before(now) {
		page.ValidationMessages = append(page.ValidationMessages, "visibility automation date must be in the future")
	} else {
		evt.VisibilityChange = visibilityChange
		evt.VisibilityChangeTime = changeTime
		evt.scheduleVisibilityChange()
		if evt.VisibilityChange != "" {
			page.ValidationMessages = append(page.ValidationMessages, fmt.Sprintf("visibility will change to '%v' in %v", evt.VisibilityChange, time.Until(changeTime).Round(time.Second)))
		}
	}

	evt.EventTitle, page.ValidationMessages = getFormValidatingLength(request, "event-title", "event title", evt.EventTitle, 100, page.ValidationMessages)
	evt.PuzzleTitle, page.ValidationMessages = getFormValidatingLength(request, "puzzle-title", "puzzle title", evt.PuzzleTitle, 100, page.ValidationMessages)
	evt.PuzzleImageURL, page.ValidationMessages = getFormValidatingLength(request, "puzzle-image", "puzzle image url", evt.PuzzleImageURL, 128, page.ValidationMessages)
	var lore string
	lore, page.ValidationMessages = getFormValidatingLength(request, "puzzle-lore", "puzzle lore", string(evt.PuzzleLore), 40000, page.ValidationMessages)
	evt.PuzzleLore = template.HTML(lore)
	files := request.MultipartForm.File["puzzle-file"]
	if len(files) > 0 {
		if files[0].Size > 10000 {
			page.ValidationMessages = append(page.ValidationMessages, "puzzle file can't be larger than 10,000 bytes")
		} else {
			file, err := files[0].Open()
			defer file.Close()
			if err != nil {
				page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open uploaded file")
			} else {
				contents, err := io.ReadAll(file)
				if err != nil {
					page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to read uploaded file")
				} else if len(puzzleName(contents)) == 0 {
					page.ValidationMessages = append(page.ValidationMessages, "file doesn't seem to be an opus magnum puzzle file")
				} else {
					evt.PuzzleFile = contents
					page.ValidationMessages = append(page.ValidationMessages, fmt.Sprintf("puzzle file '%s' uploaded!", puzzleName(contents)))
				}
			}
		}
	}
	var puzzleFileName string
	puzzleFileName, page.ValidationMessages = getFormValidatingLength(request, "puzzle-file-name", "puzzle file name", evt.PuzzleFileName, 256, page.ValidationMessages)
	if !strings.HasSuffix(puzzleFileName, ".puzzle") {
		page.ValidationMessages = append(page.ValidationMessages, "puzzle file name must end with .puzzle")
	} else {
		evt.PuzzleFileName = puzzleFileName
	}
	getCustomEvaluator := func (fieldName string) []byte {
		files := request.MultipartForm.File[fieldName]
		if len(files) == 0 {
			return nil
		}
		if files[0].Size > 64000 {
			page.ValidationMessages = append(page.ValidationMessages, "custom evaluator can't be larger than 64,000 bytes")
			return nil
		}
		file, err := files[0].Open()
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to open uploaded file")
			return nil
		}
		defer file.Close()
		contents, err := io.ReadAll(file)
		if err != nil {
			page.ValidationMessages = append(page.ValidationMessages, "internal error: unable to read uploaded file")
			return nil
		}
		// xx do more validation?
		page.ValidationMessages = append(page.ValidationMessages, "custom evaluator uploaded!")
		return contents
	}
	for i, _ := range evt.Metrics {
		m := &evt.Metrics[i]
		m.Name, page.ValidationMessages = getFormValidatingLength(request, fmt.Sprintf("metric-%d-name", i+1), fmt.Sprintf("metric %d name", i+1), m.Name, 48, page.ValidationMessages)
		m.Restrictions, page.ValidationMessages = getFormValidatingLength(request, fmt.Sprintf("metric-%d-restrictions", i+1), fmt.Sprintf("metric %d restrictions", i+1), m.Restrictions, 128, page.ValidationMessages)
		m.Submetrics[0], page.ValidationMessages = getFormValidatingLength(request, fmt.Sprintf("metric-%d-primary", i+1), fmt.Sprintf("metric %d primary", i+1), m.Submetrics[0], 128, page.ValidationMessages)
		m.Submetrics[1], page.ValidationMessages = getFormValidatingLength(request, fmt.Sprintf("metric-%d-secondary", i+1), fmt.Sprintf("metric %d secondary", i+1), m.Submetrics[1], 128, page.ValidationMessages)
		m.Submetrics[2], page.ValidationMessages = getFormValidatingLength(request, fmt.Sprintf("metric-%d-tertiary", i+1), fmt.Sprintf("metric %d tertiary", i+1), m.Submetrics[2], 128, page.ValidationMessages)
		evaluator := getCustomEvaluator(fmt.Sprintf("metric-%d-evaluator", i+1))
		if evaluator != nil {
			m.CustomEvaluatorWASM = evaluator
		} else if request.PostForm.Get(fmt.Sprintf("metric-%d-use-evaluator", i+1)) == "" {
			m.CustomEvaluatorWASM = nil
		}
	}
	if request.PostForm.Get("enable-third-metric") != "" {
		for len(evt.Metrics) < 3 {
			evt.Metrics = append(evt.Metrics, metric{ Restrictions: "default restrictions" })
		}
	} else if len(evt.Metrics) == 3 {
		evt.Metrics = evt.Metrics[:2]
	}
	evt.DisableSimulation = request.PostForm.Get("enable-cycles-validation") == ""
	evt.HidePuzzleDownloadLinkAndDisableValidation = request.PostForm.Get("hide-puzzle-download-link") != ""
	if request.PostForm.Get("display-as-post") != "" {
		evt.Style = "post"
	} else {
		evt.Style = ""
	}
	dueTime, err := time.Parse("2006-01-02@15:04", fmt.Sprintf("%s@%s", request.PostForm.Get("due-date"), request.PostForm.Get("due-time")))
	if err != nil {
		page.ValidationMessages = append(page.ValidationMessages, "invalid due date")
	} else if dueTime == evt.DueTime {
		// do nothing; the due time didn't change.
	} else if dueTime.Before(now) {
		page.ValidationMessages = append(page.ValidationMessages, "due date must be in the future")
	} else {
		evt.DueTime = dueTime
		s.mutex.Lock()
		s.eventsInDueTimeOrderNeedToBeSorted = true
		s.mutex.Unlock()
	}

	collection := collection{}
	if evt.Collection != nil {
		collection = *evt.Collection
	}
	collection.Name, page.ValidationMessages = getFormValidatingLength(request, "collection-name", "collection name", collection.Name, 48, page.ValidationMessages)
	collection.ViewKey, page.ValidationMessages = getFormValidatingLength(request, "collection-view-key", "collection id", collection.ViewKey, 48, page.ValidationMessages)
	collection.EditKey, page.ValidationMessages = getFormValidatingLength(request, "collection-edit-key", "collection secret key", collection.EditKey, 48, page.ValidationMessages)
	if collection.EditKey == "" && (collection.Name != "" || collection.ViewKey != "") {
		page.ValidationMessages = append(page.ValidationMessages, "all collections must have a secret key")
	} else {
		func () {
			s.mutex.Lock()
			defer s.mutex.Unlock()
			if !s.setCollection(evt, collection) {
				page.ValidationMessages = append(page.ValidationMessages, "collection id is already in use by another collection")
			}
		}()
	}
	evt.saveToFilesystem()
	if len(page.ValidationMessages) == 0 {
		// no need to show validation messages, so just 303 redirect to allow
		// the browser to reload the page without a warning.
		response.Header().Add("Location", request.URL.Path)
		response.WriteHeader(303)
	}
}

// this function was written by Luit van Drongelen and copied here from the x/crypto/pbkdf2 package.
func pbkdf2(password, salt []byte, iter, keyLen int, h func() hash.Hash) []byte {
	prf := hmac.New(h, password)
	hashLen := prf.Size()
	numBlocks := (keyLen + hashLen - 1) / hashLen

	var buf [4]byte
	dk := make([]byte, 0, numBlocks*hashLen)
	U := make([]byte, hashLen)
	for block := 1; block <= numBlocks; block++ {
		// N.B.: || means concatenation, ^ means XOR
		// for each block T_i = U_1 ^ U_2 ^ ... ^ U_iter
		// U_1 = PRF(password, salt || uint(i))
		prf.Reset()
		prf.Write(salt)
		buf[0] = byte(block >> 24)
		buf[1] = byte(block >> 16)
		buf[2] = byte(block >> 8)
		buf[3] = byte(block)
		prf.Write(buf[:4])
		dk = prf.Sum(dk)
		T := dk[len(dk)-hashLen:]
		copy(U, T)

		// U_n = PRF(password, U_(n-1))
		for n := 2; n <= iter; n++ {
			prf.Reset()
			prf.Write(U)
			U = U[:0]
			U = prf.Sum(U)
			for x := range U {
				T[x] ^= U[x]
			}
		}
	}
	return dk[:keyLen]
}

type parsedPuzzleAtom struct {
	AtomType byte
	U int8
	V int8
}
type parsedPuzzleBond struct {
	BondType byte
	U1 int8
	V1 int8
	U2 int8
	V2 int8
}
type parsedPuzzleMolecule struct {
	Atoms []parsedPuzzleAtom
	Bonds []parsedPuzzleBond
}
type parsedPuzzle struct {
	Valid bool
	Inputs []parsedPuzzleMolecule
	Outputs []parsedPuzzleMolecule
}
func parsePuzzleMolecule(data []byte) (parsedPuzzleMolecule, []byte) {
	var molecule parsedPuzzleMolecule
	molecule.Atoms = make([]parsedPuzzleAtom, binary.LittleEndian.Uint32(data))
	data = data[4:]
	for i := 0; i < len(molecule.Atoms); i++ {
		molecule.Atoms[i] = parsedPuzzleAtom{ data[0], int8(data[1]), int8(data[2]) }
		data = data[3:]
	}
	molecule.Bonds = make([]parsedPuzzleBond, binary.LittleEndian.Uint32(data))
	data = data[4:]
	for i := 0; i < len(molecule.Bonds); i++ {
		molecule.Bonds[i] = parsedPuzzleBond{ data[0], int8(data[1]), int8(data[2]), int8(data[3]), int8(data[4]) }
		data = data[5:]
	}
	return molecule, data
}
func parsePuzzle(data []byte) (puzzle parsedPuzzle) {
	defer func () {
		if err := recover(); err != nil {
			return
		}
	}()
	if binary.LittleEndian.Uint32(data) != 3 {
		return
	}
	n, data := parseLength(data[4:])
	data = data[n:]
	data = data[8:] // creator
	data = data[8:] // parts available
	puzzle.Inputs = make([]parsedPuzzleMolecule, binary.LittleEndian.Uint32(data))
	data = data[4:]
	for i := 0; i < len(puzzle.Inputs); i++ {
		puzzle.Inputs[i], data = parsePuzzleMolecule(data)
	}
	puzzle.Outputs = make([]parsedPuzzleMolecule, binary.LittleEndian.Uint32(data))
	data = data[4:]
	for i := 0; i < len(puzzle.Outputs); i++ {
		puzzle.Outputs[i], data = parsePuzzleMolecule(data)
	}
	puzzle.Valid = true
	return
}

type iconAttributes struct {
	svgAttributes string
	colors [2]string
	scaleViewBox bool
	atomRadius float32
	strokeWidth float32
}
var frontPageListIconAttributes = iconAttributes { " class='icon' display='none' viewBox='-4 -4 8 8' width='32' height='32'", [2]string { "", "#99a" }, false, 0.35, 0.2 }
var standaloneListIconAttributes = iconAttributes { " width='32' height='32'", [2]string { "#334", "#eee" }, true, 0.4, 0.3 }
func iconForPuzzleData(puzzleData []byte, attributes iconAttributes) template.HTML {
	puzzle := parsePuzzle(puzzleData)
	var html strings.Builder
	if !puzzle.Valid {
		viewBox := ""
		if attributes.scaleViewBox {
			viewBox = " viewBox='-2.5 -2.5 5 5'"
		}
		fmt.Fprintf(&html, "<svg xmlns='http://www.w3.org/2000/svg'%s%s>", viewBox, attributes.svgAttributes)
		if attributes.colors[0] != "" {
			fmt.Fprintf(&html, "<rect transform='rotate(45)' x='-1.25' y='-1.25' width='2.5' height='2.5' fill='%s'/>", attributes.colors[0])
		}
		fmt.Fprintf(&html, "<rect transform='rotate(45)' x='-0.75' y='-0.75' width='1.5' height='1.5' fill='%s'/>", attributes.colors[1])
		html.WriteString("</svg>")
		return template.HTML(html.String())
	}
	var bestInputOutput parsedPuzzleMolecule
	for _, input := range puzzle.Inputs {
		if len(input.Atoms) >= len(bestInputOutput.Atoms) {
			bestInputOutput = input
		}
	}
	for _, output := range puzzle.Outputs {
		if len(output.Atoms) >= len(bestInputOutput.Atoms) {
			bestInputOutput = output
		}
	}
	var minX float32
	var maxX float32
	var minY float32
	var maxY float32
	for _, atom := range bestInputOutput.Atoms {
		x := float32(atom.U) + 0.5 * float32(atom.V)
		y := -0.866 * float32(atom.V)
		if x < minX { minX = x }
		if x > maxX { maxX = x }
		if y < minY { minY = y }
		if y > maxY { maxY = y }
	}
	centerX := 0.5 * (minX + maxX)
	centerY := 0.5 * (minY + maxY)
	viewBox := ""
	if attributes.scaleViewBox {
		var viewMinX = minX - centerX - attributes.atomRadius * 2
		var viewMinY = minY - centerY - attributes.atomRadius * 2
		var viewMaxX = maxX - centerX + attributes.atomRadius * 2
		var viewMaxY = maxY - centerY + attributes.atomRadius * 2
		if viewMaxX < 1.5 {
			viewMaxX = 1.5
			viewMinX = -1.5
		}
		if viewMaxY < 1.5 {
			viewMaxY = 1.5
			viewMinY = -1.5
		}
		viewBox = fmt.Sprintf(" viewBox='%g %g %g %g'", viewMinX, viewMinY, viewMaxX - viewMinX, viewMaxY - viewMinY)
	}
	fmt.Fprintf(&html, "<svg xmlns='http://www.w3.org/2000/svg'%s%s>", viewBox, attributes.svgAttributes)
	for i := 0; i < 2; i++ {
		if attributes.colors[i] == "" {
			continue
		}
		fmt.Fprintf(&html, "<g fill='%s'>", attributes.colors[i])
		atomRadius := attributes.atomRadius
		strokeWidth := attributes.strokeWidth
		if i == 0 {
			atomRadius *= 2
			strokeWidth *= 2
		}
		for _, atom := range bestInputOutput.Atoms {
			fmt.Fprintf(&html, "<circle cx='%g' cy='%g' r='%g'/>", float32(atom.U) + 0.5 * float32(atom.V) - centerX, -0.866 * float32(atom.V) - centerY, atomRadius)
		}
		html.WriteString("</g>")
		fmt.Fprintf(&html, "<g stroke='%s' stroke-width='%g'>", attributes.colors[i], strokeWidth)
		for _, bond := range bestInputOutput.Bonds {
			fmt.Fprintf(&html, "<line x1='%g' y1='%g' x2='%g' y2='%g'/>",
				float32(bond.U1) + 0.5 * float32(bond.V1) - centerX, -0.866 * float32(bond.V1) - centerY,
				float32(bond.U2) + 0.5 * float32(bond.V2) - centerX, -0.866 * float32(bond.V2) - centerY)
		}
		html.WriteString("</g>")
	}
	html.WriteString("</svg>")
	return template.HTML(html.String())
}

func getModTime(path string) (int64, error) {
	info, err := os.Stat(filepath.Join(staticPath, path))
	if err != nil {
		return 0, err
	}
	return info.ModTime().Unix(), nil
}

func fmtResult(result interface{}) string {
	switch v := result.(type) {
	case float64:
		return strconv.FormatFloat(math.Round(v * 1000) / 1000, 'f', -1, 64)
	case map[string]interface{}:
		intervals := v["intervals"].([]interface{})
		repeatsAfter := int(v["repeatsAfter"].(float64))
		var s strings.Builder
		for i := 0; i < len(intervals); i++ {
			if i > 0 {
				s.WriteString(" ")
			}
			if i == repeatsAfter {
				s.WriteString("[")
			}
			fmt.Fprintf(&s, "%d", int(intervals[i].(float64)));
		}
		if repeatsAfter == len(intervals) {
			if len(intervals) > 0 {
				s.WriteString(" [")
			} else {
				s.WriteString("[")
			}
		}
		if repeatsAfter >= 0 {
			s.WriteString("]")
		}
		return s.String();
	case nil:
		// this is how "infinity" gets encoded.
		return "\u221e"
	default:
		return "[unknown metric data]"
	}
}

var templates = template.Must(template.New("").Funcs(template.FuncMap{
	"b64enc": base64.StdEncoding.EncodeToString,
	"lower": strings.ToLower,
	"since": time.Since,
	"modtime": getModTime,
	"fmtresult": fmtResult,
	"inc": func(i int) int { return i + 1 },
}).Parse(`
{{define "prologue"}}
	<!doctype html>
	<meta charset='utf-8'>
	<link rel='stylesheet' href='/static/style.css?{{modtime "style.css"}}'>
	<meta name='theme-color' content='#333344'>
	<meta name='viewport' content='width=800'>
{{end}}
{{define "epilogue"}}
	<script src='/static/script.js?{{modtime "script.js"}}'></script>
{{end}}
{{define "timestamp"}}<span data-timestamp='{{.Unix}}'><b>{{.Format "January 2, 2006" | lower}}</b> at <b>{{.Format "15:04"}} UTC</b></span>{{end}}
{{define "event in list"}}
	<li {{if eq .Style "post"}}class='post'{{end}}>
		<a href='/{{.ViewKey}}'>{{.Icon}}{{.EventTitle}}</a>
		{{if eq .Style "post"}}
		{{else if ge (since .DueTime) 0}}
		<span class='list-countdown'>&mdash; ended on {{template "timestamp" .DueTime}}</span>
		{{else}}
		<span class='list-countdown' data-countdown='{{.DueTime.Unix}}' data-countdown-pattern='&mdash; ends in ___' data-countdown-finished='just ended'><noscript>&mdash; ends on <b>{{.DueTime.Format "January 2, 2006" | lower}}</b> at <b>{{.DueTime.Format "15:04"}} UTC</b></noscript><b>&nbsp;</b></span>
		{{end}}
	</li>
{{end}}
{{define "subtitle-countdown"}}
	{{if ge (since .) 0}}
	<div class='event-ends-in'>event ended on {{template "timestamp" .}}</div>
	{{else}}
	<div class='event-ends-in' data-countdown='{{.Unix}}' data-countdown-pattern='ends in ___' data-countdown-finished='&lt;b>just ended&lt;/b>'><noscript>ends on <b>{{.Format "January 2, 2006" | lower}}</b> at <b>{{.Format "15:04"}} UTC</b></noscript><b>&nbsp;</b></div>
	{{end}}
{{end}}
{{define "content"}}
	{{if eq .Event.Style "post"}}
		<h1>{{.Event.EventTitle}}</h1>
		<div class='post-content'>
			{{.Event.PuzzleLore}}
		</div>
	{{else}}
		<h1>{{.Event.EventTitle}}</h1>
		{{template "subtitle-countdown" .Event.DueTime}}
		<div class='emphasized-section'>
			{{if .Event.PuzzleImageURL}}
			<img src='{{.Event.PuzzleImageURL}}'>
			{{end}}
			<h2>{{.Event.PuzzleTitle}}</h2>
			<p>{{.Event.PuzzleLore}}</p>
			{{if .Event.HidePuzzleDownloadLinkAndDisableValidation}}
			<a id='puzzle-link' style='display: none' href='data:application/octet-stream;base64,{{.Event.PuzzleFile | b64enc}}'></a>
			{{else}}
			<p>download: <a id='puzzle-link' download='{{.Event.PuzzleFileName}}' href='data:application/octet-stream;base64,{{.Event.PuzzleFile | b64enc}}'>{{.Event.PuzzleFileName}}</a> (<a href='/static/where.html'>where do i put this?</a>)</p>
			{{end}}
		</div>
		<h2>submission form</h2>
		{{if ge (since .Event.DueTime) 0}}
			<div class='emphasized-section'>this event has ended &mdash; only showcase submissions will be accepted</div>
		{{else}}
			to be accepted, solutions must be submitted by {{template "timestamp" .Event.DueTime}}<br>
			<span data-countdown='{{.Event.DueTime.Unix}}' data-countdown-pattern='that&#39;s ___ from now&lt;br>' data-countdown-finished='&lt;b>time&#39;s up!&lt;/b>&lt;br>'></span><br>
		{{end}}
		<form method='post' enctype='multipart/form-data' action='/{{.Event.ViewKey}}' id='submission-form' data-metrics='{{.Event.MetricsJSON | printf "%s"}}'{{if .Event.HidePuzzleDownloadLinkAndDisableValidation}} data-disable-puzzle-file-name-validation{{end}}>
			by what name would you like to be known?<br>
			<input type='text' name='submitter' placeholder='submitter name' value='{{.SavedSubmitter}}'><input type='text' name='pronouns' placeholder='pronouns' value='{{.SavedPronouns}}'><br>
			<br>
			how can you be contacted? (fill in either or both)<br>
			<input type='text' name='discord' placeholder='discord tag' value='{{.SavedDiscord}}'><br>
			<input type='email' name='email' placeholder='email address' value='{{.SavedEmail}}'><br>
			<br>
			entering a secret key (8+ characters) lets you see your solutions and add notes after submission<br>
			<input type='password' name='secret-key' placeholder='secret key' value='{{.SavedSecretKey}}'><br>
			<br>
			<label><input type='checkbox' name='save'{{if .SavedSubmitter}} checked{{end}}> remember this information in your browser so you don't have to type it again</label><br><br>
			which .solution files would you like to submit? (<a href='/static/where.html'>where do i find these?</a>)<br>
			<br>
			<div class='submission active' id='submission-upload-box'>
				<input type='file' name='solution' multiple><br>
				<label><input type='checkbox' id='showcase-checkbox' name='showcase'{{if ge (since .Event.DueTime) 0}} checked disabled{{end}}> non-scoring showcase</label>
				{{if ge (since .Event.DueTime) 0}}<input type='hidden' name='showcase' value='after-due-date'>{{end}}
			</div>
			<input type='submit' value='submit'> <input type='submit' name='show-previous' value='show previous submissions'>
			<div id='previous'>
				{{if .ShowingPreviousSubmissions}}
				<h2>previous submissions</h2>
				{{ $timestamp := 0 }}
				{{range .Submissions}}
				{{if ne .Metadata.SubmissionTime.Unix $timestamp}}
				{{ $timestamp = .Metadata.SubmissionTime.Unix }}
				<span data-timestamp='{{.Metadata.SubmissionTime.Unix}}'><b>{{.Metadata.SubmissionTime.Format "January 2, 2006" | lower}}</b> at <b>{{.Metadata.SubmissionTime.Format "15:04"}} UTC</b></span><br><br>
				{{end}}
				<div class='submission previous'>
					<div class='info'>
						{{template "previous-submission-info" .}}
					</div>
					{{if .Metadata.Key}}
					<div class='buttons'>
						<button name='delete-previous-submission' value='{{.Metadata.Key}}'>delete</button>
						<button name='download-previous-submission' value='{{.Metadata.Key}}'>download</button>
						<button name='notes-for-solution' value='{{.Metadata.Key}}'>
							{{if eq .NotesWordCount 0}}
							add notes
							{{else if eq .NotesWordCount 1}}
							edit notes (1 word)
							{{else}}
							edit notes ({{.NotesWordCount}} words)
							{{end}}
						</button>
					</div>
					{{end}}
				</div>
				{{else}}
				no previous submissions found with this key.
				{{end}}
				{{end}}
			</div>
		</form>
	{{end}}
{{end}}
{{define "previous-submission-info"}}
	<b>{{.Metadata.TruncatedSolutionName}}</b>
	{{if .Metadata.IsShowcase}}
	(showcase)
	{{else}}
	<div class='metrics'>
		{{ $metrics := .Event.Metrics }}
		{{range $i, $m := .Metadata.EvaluatedMetrics}}
		{{if .}}
		{{if $i}}<br>{{end}}
		{{if lt $i (len $metrics)}}<b>{{(index $metrics $i).Name}}</b>{{end}}
		{{range $j, $v := $m}}{{if $j}}/{{end}}<span class='metric{{inc $i}}-{{$j}}'>{{$v | fmtresult}}</span>{{end}}
		{{end}}
		{{end}}
	</div>
	{{end}}
{{end}}
{{define "edit notes"}}
	{{template "prologue"}}
	<title>editing notes for '{{.Metadata.TruncatedSolutionName}}'</title>
	<div id='container'>
		{{template "breadcrumbs" .Event}}
		<h1>{{.Event.EventTitle}}</h1>
		{{template "subtitle-countdown" .Event.DueTime}}
		<h2>editing notes for submission</h2>
		<form id='edit-notes-form' method='post' enctype='multipart/form-data' action='/{{.Event.ViewKey}}'>
			<div class='submission previous'>
				<div class='info'>
					{{template "previous-submission-info" .}}
				</div>
				<div class='buttons'>
					<button id='notes-discard-changes' form='discard-form'>discard changes</button>
					<button id='notes-save-changes' name='notes-for-solution' value='{{.Metadata.Key}}'>save and finish editing</button>
				</div>
			</div>
			<textarea id='edit-notes' name='notes'>{{.Notes}}</textarea>
		</form>
		<form id='discard-form' method='get' action='/{{.Event.ViewKey}}'></form>
	</div>
	{{template "epilogue"}}
{{end}}
{{define "delete submission"}}
	{{template "prologue"}}
	<title>editing notes for '{{.Metadata.TruncatedSolutionName}}'</title>
	<div id='container'>
		{{template "breadcrumbs" .Event}}
		<h1>{{.Event.EventTitle}}</h1>
		{{template "subtitle-countdown" .Event.DueTime}}
		<h2>deleting {{.Metadata.TruncatedSolutionName}}</h2>
		<form id='delete-submission-form' enctype='multipart/form-data' method='post' action='/{{.Event.ViewKey}}'>
			<div class='submission previous'>
				<div class='info'>
					{{template "previous-submission-info" .}}
				</div>
			</div>
			<p>are you sure you want to delete this submission?</p>
			<button name='really-delete-previous-submission' value='{{.Metadata.Key}}'>delete it</button>
			<input type='submit' formmethod='get' value='do not'>
		</form>
	</div>
	{{template "epilogue"}}
{{end}}
{{define "breadcrumbs"}}
	<header><a href='/'><b>events.critelli.technology</b></a> /
	{{if .Collection}}
	{{if .Collection.ViewKey}}
	{{if .Collection.Name}}
	<a href='/collection/{{.Collection.ViewKey}}'>{{.Collection.Name}}</a> /
	{{else}}
	<a href='/collection/{{.Collection.ViewKey}}'>untitled collection</a> /
	{{end}}
	{{else}}
	{{if .Collection.Name}}
	{{.Collection.Name}} /
	{{end}}
	{{end}}
	{{end}}</header>
{{end}}
{{define "message list"}}
	{{if .}}
	<div class='emphasized-section'>
		{{if eq (len .) 1}}
		{{index . 0}}
		{{else}}
		{{range .}}
		<li>{{.}}</li>
		{{end}}
		{{end}}
	</div>
	{{end}}
{{end}}
{{define "front page"}}
	{{template "prologue"}}
	<title>events.critelli.technology</title>
	<div id='container'>
		<header id='front-header'>
			<h1>events.critelli.technology</h1>
		</header>
		<h2>ongoing</h2>
		{{range .OngoingEventsByCollection}}
		{{if (index . 0).Collection.ViewKey}}
		<h3><a href='/collection/{{(index . 0).Collection.ViewKey}}'>{{(index . 0).Collection.Name}}</a></h3>
		{{else if (index . 0).Collection.Name}}
		<h3>{{(index . 0).Collection.Name}}</h3>
		{{end}}
		<ul class='event-list'>
			{{range .}}
			{{template "event in list" .}}
			{{end}}
		</ul>
		{{else}}
		<ul class='event-list'>
			<li><i>no ongoing events</i></li>
		</ul>
		{{end}}
		<hr>
		<h2>completed</h2>
		{{ $completed := .CompletedEventsByCollection }}
		{{range .Collections}}
		{{if (index $completed .)}}
		{{if .ViewKey}}
		<h3><a href='/collection/{{.ViewKey}}'>{{.Name}}</a></h3>
		{{else if .Name}}
		<h3>{{.Name}}</h3>
		{{end}}
		<ul class='event-list'>
			{{range (index $completed .)}}
			{{template "event in list" .}}
			{{end}}
		</ul>
		{{end}}
		{{end}}
		<hr>
		<p>
			<form method='post' action='/new'>
				<input type='submit' value='create new event'>
			</form>
		</p>
	</div>
	{{template "epilogue"}}
{{end}}
{{define "collection"}}
	{{template "prologue"}}

	{{ $firstActive := false }}
	{{ $hasCompleted := false }}
	{{range .}}
	{{if ge (since .DueTime) 0}}
	{{ $hasCompleted = true }}
	{{else if (not $firstActive)}}
	{{ $firstActive = . }}
	{{end}}
	{{end}}
	{{if (not $firstActive)}}
	{{ $firstActive = (index .0) }}
	{{end}}

	{{if (index . 0).Collection.Name}}
	<meta content='{{(index . 0).Collection.Name}}' property='og:title'>
	{{if $firstActive }}
	<meta content='{{$firstActive.PuzzleImageURL}}' property='og:image'>
	{{end}}
	<meta content='view puzzles and submit solutions' property='og:description'>
	<title>{{(index . 0).Collection.Name}}</title>
	{{end}}
	<div id='container'>
		<header><a href='/'><b>events.critelli.technology</b></a> /</header>
		{{if (index . 0).Collection.Name}}
		<h1>{{(index . 0).Collection.Name}}</h1>
		{{end}}
		{{if $firstActive}}
		<h2>ongoing</h2>
		<ul class='event-list'>
			{{range .}}
			{{if lt (since .DueTime) 0}}
			{{template "event in list" .}}
			{{end}}
			{{end}}
		</ul>
		{{end}}
		{{if $hasCompleted}}
		<h2>completed</h2>
		<ul class='event-list'>
			{{range .}}
			{{if ge (since .DueTime) 0}}
			{{template "event in list" .}}
			{{end}}
			{{end}}
		</ul>
		{{end}}
	</div>
	{{template "epilogue"}}
{{end}}
{{define "view"}}
	{{template "prologue"}}
	<meta content='{{.Event.EventTitle}}' property='og:site_name'>
	{{if ne .Event.Style "post"}}
	<meta content='{{.Event.PuzzleTitle}}' property='og:title'>
	{{end}}
	<meta content='{{.Event.PuzzleLore}}' property='og:description'>
	{{if .Event.PuzzleImageURL}}
	<meta name='twitter:card' content='summary_large_image'>
	<meta content='{{.Event.PuzzleImageURL}}' property='og:image'>
	{{end}}
	<link rel='icon' href='/icon/{{.Event.ViewKey}}'>
	<link rel='mask-icon' href='/icon/{{.Event.ViewKey}}' color='#334'>
	<title>{{.Event.EventTitle}}</title>
	<div id='container'>
		{{if eq .Event.Visibility "public"}}
		{{template "breadcrumbs" .Event}}
		{{end}}
		{{template "message list" .ValidationMessages}}
		{{template "content" .}}
	</div>
	{{template "epilogue"}}
{{end}}
{{define "private"}}
	{{template "prologue"}}
	<title>upcoming event</title>
	<div id='container'>
		this event is not yet accepting submissions.
	</div>
	{{template "epilogue"}}
{{end}}
{{define "empty-collection"}}
	{{template "prologue"}}
	<title>empty collection</title>
	<div id='container'>
		<header><a href='/'><b>events.critelli.technology</b></a> /</header>
		<h1>empty collection</h1>
		<ul class='event-list'>
			<li><i>no events</i></li>
		</ul>
	</div>
	{{template "epilogue"}}
{{end}}
{{define "event"}}
	{{template "prologue"}}
	<title>admin page for '{{.Event.EventTitle}}'</title>
	<div id='stretch-container'>
		{{template "breadcrumbs" .Event}}
		{{template "message list" .ValidationMessages}}
		<h1>{{.Event.EventTitle}}</h1>
		{{template "subtitle-countdown" .Event.DueTime}}
		<ul>
			<li><a href='/edit/{{.Event.EditKey}}'>edit this event</a></li>
			<li>share submission link: <a href='/{{.Event.ViewKey}}'><!--xx url base-->http://events.critelli.technology/{{.Event.ViewKey}}</a>
			<li><a href='/delete/{{.Event.EditKey}}'>delete this event</a></li>
		</ul>
		<p>be sure to save the <a href='/event/{{.Event.EditKey}}'>current url</a> -- you'll need it if you want to access these links later!</p>
		<h2>submissions</h1>
		{{ $empty := true }}{{range .Submissions}}{{if not .Metadata.IsDeleted}}{{ $empty = false }}{{end}}{{end}}
		{{if $empty }}
		<i>no submissions yet.</i>
		{{else}}
		<form id='table-form' method='post' action='/event/{{.Event.EditKey}}' data-puzzle='data:application/octet-stream;base64,{{.Event.PuzzleFile | b64enc}}'>
			<table class='submissions'>
				<thead>
					<tr>
						<th>submitter</th>
						<th>pronouns</th>
						<th>discord</th>
						<th>email</th>
						<th>secret key hash</th>
						<th>solution name</th>
						{{range .Event.Metrics}}
						<th>{{.Name}}</th>
						{{end}}
						<th>submission time (UTC)</th>
						<th>showcase?</th>
						<th>download</th>
						<th>notes</th>
						<th>delete?</th>
					</tr>
				</thead>
				{{range $i, $s := .Submissions}}
				{{if not .Metadata.IsDeleted}}
				<tr>
					<td><input type='text' name='{{$i}}.name' size='{{len .Metadata.SubmitterName}}' value='{{.Metadata.SubmitterName}}'></td>
					<td><input type='text' name='{{$i}}.pronouns' size='{{if eq (len .Metadata.SubmitterPronouns) 0}}9{{else}}{{len .Metadata.SubmitterPronouns}}{{end}}' value='{{.Metadata.SubmitterPronouns}}'></td>
					<td><input type='text' name='{{$i}}.discord' size='{{len .Metadata.SubmitterDiscord}}' value='{{.Metadata.SubmitterDiscord}}'></td>
					<td><input type='text' name='{{$i}}.email' size='{{len .Metadata.SubmitterEmail}}' value='{{.Metadata.SubmitterEmail}}'></td>
					<td title='{{.Metadata.SecretKeyHashFull}}'>{{.Metadata.SecretKeyHashTruncated}}</td>
					<td>{{.Metadata.TruncatedSolutionName}}</td>
					{{range $j, $m := .Event.Metrics}}
					{{if lt $j (len $s.Metadata.EvaluatedMetrics)}}
					<td>{{range $k, $v := (index $s.Metadata.EvaluatedMetrics $j)}}{{if $k}}/{{end}}{{$v | fmtresult}}{{end}}</td>
					{{else}}
					<td></td>
					{{end}}
					{{end}}
					<td data-sort='{{.Metadata.SubmissionTime.UnixNano}}'>{{.Metadata.SubmissionTime.Format "January 2, 2006 at 15:04" | lower}}</td>
					<td><input type='checkbox' name='{{$i}}.showcase'{{if .Metadata.IsShowcase}} checked{{end}}></td>
					<td><a class='solution-file' download='{{.Metadata.SolutionFileName .Event.PuzzleTitle}}' href='/submission/{{.Event.SubmissionsKey}}?offset={{.Metadata.ByteOffsetInFile}}&length={{.Metadata.ByteLengthInFile}}'>solution file</a></td>
					{{if eq .NotesWordCount 0}}
					<td></td>
					{{else if eq .NotesWordCount 1}}
					<td><a href='/notes/{{.Event.SubmissionsKey}}?submission={{.Metadata.Key}}'>show (1 word)</a></td>
					{{else}}
					<td><a href='/notes/{{.Event.SubmissionsKey}}?submission={{.Metadata.Key}}'>show ({{.NotesWordCount}} words)</a></td>
					{{end}}
					<td><input type='checkbox' name='{{$i}}.deleted'{{if .Metadata.IsDeleted}} checked{{end}}></td>
				</tr>
				{{end}}
				{{end}}
			</table>
			<p class='wrap-to-viewport'><b>**unique submitters ({{len .UniqueSubmitters}}):**</b> {{range $i, $s := .UniqueSubmitters}}{{if ne $i 0}}, {{end}}{{$s}}{{end}}</p>
			<p>download all submissions as a zip file: <a download='{{.Event.ViewKey}}.zip' href='/archive/{{.Event.SubmissionsKey}}'>{{.Event.ViewKey}}.zip</a></p>
			<p>to import this table into google sheets, use<br><tt>=IMPORTDATA("<a href='/csv/{{.Event.SubmissionsKey}}'>http://events.critelli.technology/csv/{{.Event.SubmissionsKey}}</a>",",","en_US")</tt></p>
			<input type='submit' value='save changes'> <input type='reset' value='reset form'>
		</form>
		{{end}}
	</div>
	{{template "epilogue"}}
{{end}}
{{define "edit"}}
	{{template "prologue"}}
	<title>editing '{{.Event.EventTitle}}'</title>
	<div id='editor'>
		<a href='/event/{{.Event.EditKey}}'>back to event page</a>
		{{template "message list" .ValidationMessages}}
		<form method='post' enctype='multipart/form-data' action='/edit/{{.Event.EditKey}}' id='edit-form'>
			<h1>edit event</h1>
			event visibility<br>
			<label class='radio-flex'><input type='radio' name='visibility' value='private'{{if eq .Event.Visibility "private"}} checked{{end}}> <div><b>private</b> - event can't be seen, even with the direct submission link</div></label>
			<label class='radio-flex'><input type='radio' name='visibility' value='unlisted'{{if eq .Event.Visibility "unlisted"}} checked{{end}}> <div><b>unlisted</b> - event can be seen with a direct link, but does not appear on the front page</div></label>
			<label class='radio-flex'><input type='radio' name='visibility' value='public'{{if eq .Event.Visibility "public"}} checked{{end}}> <div><b>public</b> - event appears on the <a href='/'>front page</a></div></label><br>
			<details>
				<summary>event visibility automation (aka event start date)</summary><br>
				on date (YYYY-MM-DD or use built-in browser controls)<br>
				<input type='date' name='visibility-change-date' value='{{.Event.VisibilityChangeTime.Year | printf "%04d"}}-{{.Event.VisibilityChangeTime.Month | printf "%02d"}}-{{.Event.VisibilityChangeTime.Day | printf "%02d"}}'><br><br>
				at time (hh:mm) in UTC<br>
				<input type='time' name='visibility-change-time' value='{{.Event.VisibilityChangeTime.Hour | printf "%02d"}}:{{.Event.VisibilityChangeTime.Minute | printf "%02d"}}'><br><br>
				(<span id='visibility-change-date-preview'></span>)<br><br>
				<label class='radio-flex'><input type='radio' name='visibility-change' value=''{{if eq .Event.VisibilityChange ""}} checked{{end}}> <div><b>do nothing</b></div></label>
				<label class='radio-flex'><input type='radio' name='visibility-change' value='private'{{if eq .Event.VisibilityChange "private"}} checked{{end}}> <div><b>become private</b> - event can't be seen, even with the direct submission link</div></label>
				<label class='radio-flex'><input type='radio' name='visibility-change' value='unlisted'{{if eq .Event.VisibilityChange "unlisted"}} checked{{end}}> <div><b>become unlisted</b> - event can be seen with a direct link, but does not appear on the front page</div></label>
				<label class='radio-flex'><input type='radio' name='visibility-change' value='public'{{if eq .Event.VisibilityChange "public"}} checked{{end}}> <div><b>become public</b> - event appears on the <a href='/'>front page</a></div></label>
			</details><br>
			event title<br>
			<input type='text' name='event-title' value='{{.Event.EventTitle}}'><br><br>
			<details>
				<summary>is this event part of a collection?</summary><br>
				events with the same secret key are grouped on the front page (if the collection has a name) and appear on their own collection page (if the collection has an id)<br><br>
				secret collection key<br>
				<input type='text' name='collection-edit-key' value='{{if .Event.Collection}}{{.Event.Collection.EditKey}}{{end}}'><br><br>
				collection name<br>
				<input type='text' name='collection-name' value='{{if .Event.Collection}}{{.Event.Collection.Name}}{{end}}'><br><br>
				collection id (for url)<br>
				<input type='text' name='collection-view-key' value='{{if .Event.Collection}}{{.Event.Collection.ViewKey}}{{end}}'><br><br>
			</details><br>
			puzzle file (leave blank to keep puzzle file unchanged)<br>
			<input type='file' name='puzzle-file'><br><br>
			puzzle file name<br>
			<input type='text' spellcheck='false' name='puzzle-file-name' value='{{.Event.PuzzleFileName}}'><br><br>
			puzzle title<br>
			<input type='text' name='puzzle-title' value='{{.Event.PuzzleTitle}}'><br><br>
			puzzle image url<br>
			<input type='text' spellcheck='false' name='puzzle-image' value='{{.Event.PuzzleImageURL}}'><br><br>
			puzzle description and lore (html is allowed, have fun)<br>
			<textarea class='puzzle-lore' name='puzzle-lore'>{{.Event.PuzzleLore | html}}</textarea><br><br>
			event id (for url)<br>
			<input type='text' name='view-key' value='{{.Event.ViewKey}}'><br><br>
			due date (YYYY-MM-DD or use built-in browser controls)<br>
			<input type='date' name='due-date' value='{{.Event.DueTime.Year | printf "%04d"}}-{{.Event.DueTime.Month | printf "%02d"}}-{{.Event.DueTime.Day | printf "%02d"}}'><br><br>
			due time (hh:mm) in UTC<br>
			<input type='time' name='due-time' value='{{.Event.DueTime.Hour | printf "%02d"}}:{{.Event.DueTime.Minute | printf "%02d"}}'><br><br>
			<div id='due-date-preview'></div>
			<h2>metric evaluation section</h2>
			test solution for metric evaluation (not sent to server; metrics will be evaluated for this solution as you type):<br>
			<input type='file' id='test-solution' data-puzzle='data:application/octet-stream;base64,{{.Event.PuzzleFile | b64enc}}'>
			{{range $i, $m := .Event.Metrics}}
			{{ $num := inc $i }}
			<h3>metric {{$num}}</h3>
			<p id='metric-{{$num}}-errors'><i>choose solution file above to test metric evaluation</i></p>
			metric {{$num}} name<br>
			<input type='text' spellcheck='false' name='metric-{{$num}}-name' value='{{.Name}}'><br><br>
			metric {{$num}} primary (<a target='_blank' href='/static/metrics.html'>metrics reference</a>)<span id='metric-{{$num}}-primary-value'></span><br>
			<input type='text' spellcheck='false' id='metric-{{$num}}-primary' name='metric-{{$num}}-primary' value='{{index .Submetrics 0}}'><br><br>
			metric {{$num}} secondary<span id='metric-{{$num}}-secondary-value'></span><br>
			<input type='text' spellcheck='false' id='metric-{{$num}}-secondary' name='metric-{{$num}}-secondary' value='{{index .Submetrics 1}}'><br><br>
			metric {{$num}} tertiary<span id='metric-{{$num}}-tertiary-value'></span><br>
			<input type='text' spellcheck='false' id='metric-{{$num}}-tertiary' name='metric-{{$num}}-tertiary' value='{{index .Submetrics 2}}'><br><br>
			metric {{$num}} restrictions<br>
			<input type='text' spellcheck='false' id='metric-{{$num}}-restrictions' name='metric-{{$num}}-restrictions' value='{{.Restrictions}}'><br><br>
			<label><input type='checkbox' name='metric-{{$num}}-use-evaluator'{{if .CustomEvaluatorWASM}} checked{{end}}> use custom evaluator for metric {{$num}} (for computation puzzles)</label>:<br>
			<input type='file' name='metric-{{$num}}-evaluator'><br><br>
			{{end}}
			<br><label><input type='checkbox' name='enable-third-metric'{{if eq (len .Event.Metrics) 3}} checked{{end}}> enable metric 3 (save changes to show the fields)</label><br><br>
			<br><label><input type='checkbox' name='display-as-post'{{if eq .Event.Style "post"}} checked{{end}}> display as a "post" (will disable submissions)</label><br><br>
			<br><label><input type='checkbox' name='enable-cycles-validation'{{if not .Event.DisableSimulation}} checked{{end}}> validate submission cycle counts and area using simulation</label><br><br>
			<br><label><input type='checkbox' name='hide-puzzle-download-link'{{if .Event.HidePuzzleDownloadLinkAndDisableValidation}} checked{{end}}> hide the link to download the puzzle file and disable puzzle file name validation (for dynamically generated puzzles)</label><br><br>
			<div id='save-changes'><input type='submit' value='save changes'><input type='reset' value='reset form'></div>
		</form>
	</div>
	<div id='container'>
		submission link to share:<br>
		<a href='/{{.Event.ViewKey}}'><!--xx url base-->http://events.critelli.technology/{{.Event.ViewKey}}</a><br><br>
		below is a preview of the submission page - press "save changes" to update it<br><br>
		{{template "content" .}}
	</div>
	{{template "epilogue"}}
{{end}}
{{define "delete"}}
	{{template "prologue"}}
	<title>deleting '{{.Event.EventTitle}}'</title>
	<div id='container'>
		<a href='/event/{{.Event.EditKey}}'>back to event page</a>
		<h1>deleting &ldquo;{{.Event.EventTitle}}&rdquo;</h1>
		<p>are you sure you want to delete this event?</p>
		<form method='post' action='/delete/{{.Event.EditKey}}'>
		<input type='submit' value='delete it'>
		<input type='submit' formaction='/event/{{.Event.EditKey}}' formmethod='get' value='do not'>
		</form>
	</div>
	{{template "epilogue"}}
{{end}}
{{define "deleted"}}
	{{template "prologue"}}
	<title>delete '{{.Event.EventTitle}}'</title>
	<div id='container'>
		<a href='/'>back to front page</a>
		<h1>deleting &ldquo;{{.Event.EventTitle}}&rdquo;</h1>
		<p>this event has been deleted.</p>
	</div>
	{{template "epilogue"}}
{{end}}
`))
