'use strict';
let puzzleData;
function localizeDatetimes() {
    let datetimes = document.querySelectorAll('[data-timestamp]');
    for (let i = 0; i < datetimes.length; ++i) {
        let date = new Date(datetimes[i].dataset.timestamp * 1000);
        let month = [
            'january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september', 'october', 'november', 'december'
        ][date.getMonth()];
        // https://stackoverflow.com/a/68065499
        let timezone = new Date().toLocaleDateString(undefined, {day:'2-digit', timeZoneName: 'short' }).substring(4);
        datetimes[i].innerHTML = `<b>${month} ${date.getDate()}, ${date.getFullYear()}</b> at <b>${date.getHours()}:${('0' + date.getMinutes()).slice(-2)} ${timezone}</b>`;
    }
}
localizeDatetimes();
function tick() {
    let countdowns = document.querySelectorAll('[data-countdown]');
    for (let i = 0; i < countdowns.length; ++i) {
        let date = new Date(countdowns[i].dataset.countdown * 1000);
        let duration = [
            { seconds: 60 * 60 * 24, singular: 'day', plural: 'days' },
            { seconds: 60 * 60, singular: 'hour', plural: 'hours' },
            { seconds: 60, singular: 'minute', plural: 'minutes' },
            { seconds: 1, singular: 'second', plural: 'seconds' },
        ];
        let secondsUntil = (date - new Date()) / 1000;
        let countdown = '';
        for (let j = 0; j < duration.length; ++j) {
            let t = Math.floor(secondsUntil / duration[j].seconds);
            if (t <= 0)
                continue;
            if (countdown.length > 0)
                countdown += ' and ';
            else if (duration.length > j + 2)
                duration.length = j + 2;
            countdown += t + ' ' + (t === 1 ? duration[j].singular : duration[j].plural);
            secondsUntil -= t * duration[j].seconds;
        }
        if (countdown === '')
            countdowns[i].innerHTML = countdowns[i].dataset.countdownFinished;
        else
            countdowns[i].innerHTML = countdowns[i].dataset.countdownPattern.replace('___', `<b>${countdown}</b>`);
    }
}
let countdownsInterval;
function updateCountdowns() {
    if (countdownsInterval)
        clearInterval(countdownsInterval);
    if (document.querySelectorAll('[data-countdown]').length > 0) {
        tick();
        countdownsInterval = setInterval(tick, 1000);
    }
}
updateCountdowns();
let editForm = document.getElementById('edit-form');
if (editForm) {
    let elems = editForm.elements;
    for (let i = 0; i < elems.length; ++i) {
        if (!elems[i].name)
            continue;
        elems[i].initialValue = elementValue(elems[i]);
        elems[i].addEventListener('input', function (e) {
            checkDirty();
        });
    }
    editForm.addEventListener('reset', function (e) {
        setTimeout(checkDirty, 0);
    });
    editForm.elements['puzzle-file'].addEventListener('change', function (e) {
        let reader = new FileReader();
        reader.onload = function (e) {
            let data = new Uint8Array(reader.result);
            if (data[0] !== 3)
                return;
            puzzleData = data;
            editForm.elements['puzzle-title'].value = new TextDecoder().decode(data.slice(5, 5 + data[4]));
        };
        reader.readAsArrayBuffer(this.files[0]);
        if (this.files[0].name)
            editForm.elements['puzzle-file-name'].value = this.files[0].name;
    });
    editForm.elements['collection-edit-key'].addEventListener('input', function (e) {
        let editKey = this.value;
        console.log('/collection-json/' + editKey);
        fetch('/collection-json/' + this.value).then(function (response) {
            if (!response.ok)
                return {};
            return response.json();
        }).then(function (data) {
            if (editForm.elements['collection-edit-key'].value !== data.EditKey)
                return;
            editForm.elements['collection-name'].value = data.Name;
            editForm.elements['collection-view-key'].value = data.ViewKey;
        });
    });
    function configureDatePreview(name) {
        function update() {
            let date = new Date(`${editForm.elements[`${name}-date`].value}T${editForm.elements[`${name}-time`].value}Z`);
            document.getElementById(`${name}-date-preview`).innerHTML = `in your local time zone, that's <span data-timestamp='${Math.floor(date.getTime() / 1000)}'></span>`;
            localizeDatetimes();
        }
        editForm.elements[`${name}-date`].addEventListener('input', update);
        editForm.elements[`${name}-time`].addEventListener('input', update);
        update();
    }
    configureDatePreview('due');
    configureDatePreview('visibility-change');
}
function elementValue(elem) {
    if (elem.type === 'checkbox' || elem.type === 'radio')
        return elem.checked ? 'checked' : 'unchecked';
    else
        return elem.value;
}
function checkDirty() {
    let dirty = false;
    let elems = editForm.elements;
    for (let i = 0; i < elems.length; ++i) {
        if (elems[i].name && elementValue(elems[i]) !== elems[i].initialValue)
            dirty = true;
    }
    document.getElementById('save-changes').className = dirty ? 'sticky' : '';
}
function evaluateMetrics(puzzle, solution, metrics, callback, evaluatorMessageCallback) {
    let worker = new Worker('/static/verify.js?' + Date.now());
    let timer;
    function cancel() {
        if (timer)
            clearTimeout(timer);
        timer = null;
        if (worker)
            worker.terminate();
        worker = null;
    }
    function reset() {
        if (timer)
            clearTimeout(timer);
        timer = setTimeout(function () {
            cancel();
            callback({ results: null, error: 'time limit reached while evaluating metrics' });
        }, 60000);
    }
    reset();
    worker.addEventListener('message', function (e) {
        if (e.data.evaluatorMessage || e.data.evaluatorMessageAttribute) {
            reset();
            if (evaluatorMessageCallback)
                evaluatorMessageCallback(e.data.targetSelector, e.data.evaluatorMessage, e.data.evaluatorMessageAttribute, e.data.evaluatorMessageValue);
            return;
        }
        cancel();
        callback(e.data);
    });
    worker.addEventListener('error', function (e) {
        cancel();
        callback({ results: null, error: 'internal error: ' + e.message + ' (' + e.filename + ':' + e.lineno + ')'});
    });
    worker.postMessage({ puzzle: puzzle, solution: solution, metrics: metrics, supportsEvaluatorMessages: !!evaluatorMessageCallback })
    return cancel;
}
function formatResult(result) {
    if (typeof result === 'number')
        return Math.round(result * 1000) / 1000;
    else if (typeof result === 'object' && Array.isArray(result.intervals)) {
        let s = '';
        for (let i = 0; i < result.intervals.length; ++i) {
            if (i > 0)
                s += ' ';
            if (i === result.repeatsAfter)
                s += '[';
            s += result.intervals[i];
        }
        if (result.repeatsAfter === result.intervals.length) {
            if (result.intervals.length > 0)
                s += ' [';
            else
                s += '[';
        }
        if (result.repeatsAfter >= 0)
            s += ']';
        return s;
    }
}
let testSolution = document.getElementById('test-solution');
if (testSolution) {
    let solutionData;
    let cancelEvaluation;
    function testMetrics() {
        if (!puzzleData || !solutionData)
            return;
        if (cancelEvaluation)
            cancelEvaluation();
        cancelEvaluation = evaluateMetrics(puzzleData, solutionData, {
            Metrics: [
                {
                    Name: editForm.elements['metric-1-name'].value,
                    Restrictions: document.getElementById('metric-1-restrictions').value,
                    Submetrics: [
                        document.getElementById('metric-1-primary').value,
                        document.getElementById('metric-1-secondary').value,
                        document.getElementById('metric-1-tertiary').value,
                    ]
                },
                {
                    Name: editForm.elements['metric-2-name'].value,
                    Restrictions: document.getElementById('metric-2-restrictions').value,
                    Submetrics: [
                        document.getElementById('metric-2-primary').value,
                        document.getElementById('metric-2-secondary').value,
                        document.getElementById('metric-2-tertiary').value,
                    ]
                },
            ],
        }, function (result) {
            document.getElementById('metric-1-errors').textContent = result.metricErrors[0] || result.error || 'no errors';
            document.getElementById('metric-2-errors').textContent = result.metricErrors[1] || result.error || 'no errors';
            function showValue(id, i, j) {
                if (result && result.results[i] && typeof result.results[i][j] !== 'undefined')
                    document.getElementById(id).textContent = ' = ' + formatResult(result.results[i][j]);
                else
                    document.getElementById(id).textContent = '';
            }
            showValue('metric-1-primary-value', 0, 0);
            showValue('metric-1-secondary-value', 0, 1);
            showValue('metric-1-tertiary-value', 0, 2);
            showValue('metric-2-primary-value', 1, 0);
            showValue('metric-2-secondary-value', 1, 1);
            showValue('metric-2-tertiary-value', 1, 2);
        });
    }
    document.getElementById('metric-1-primary').oninput = testMetrics;
    document.getElementById('metric-1-secondary').oninput = testMetrics;
    document.getElementById('metric-1-tertiary').oninput = testMetrics;
    document.getElementById('metric-1-restrictions').oninput = testMetrics;
    document.getElementById('metric-2-primary').oninput = testMetrics;
    document.getElementById('metric-2-secondary').oninput = testMetrics;
    document.getElementById('metric-2-tertiary').oninput = testMetrics;
    document.getElementById('metric-2-restrictions').oninput = testMetrics;
    fetch(testSolution.dataset.puzzle).then(function (response) {
        return response.arrayBuffer();
    }).then(function (data) {
        puzzleData = new Uint8Array(data);
    }).then(testMetrics);
    testSolution.addEventListener('change', function (e) {
        let reader = new FileReader();
        reader.onload = function (e) {
            solutionData = new Uint8Array(reader.result);
            testMetrics();
        };
        reader.readAsArrayBuffer(this.files[0]);
    });
}
function scrollHeight() {
    return Math.max(
        document.body.scrollHeight, document.documentElement.scrollHeight,
        document.body.offsetHeight, document.documentElement.offsetHeight,
        document.body.clientHeight, document.documentElement.clientHeight
    );
}
function beforeSubmit() {
    window.onbeforeunload = null;
}
function renderNotesButton(submission, solutionName, results, existingNotes, index) {
    let notesButton = submission.querySelector('.notes-button');
    let notesContent = document.createElement('input');
    notesContent.type = 'hidden';
    notesContent.name = `notes-content-${index}`;
    submission.appendChild(notesContent);
    function updateNotesButtonWithNotes(notes) {
        notesContent.value = notes;
        let trimmed = notes.trim();
        let words = trimmed.length === 0 ? 0 : trimmed.split(/\s+/u).length;
        if (words === 0)
            notesButton.textContent = 'add notes';
        else if (words === 1)
            notesButton.textContent = 'edit notes (1 word)';
        else
            notesButton.textContent = `edit notes (${words} words)`;
    }
    notesButton.addEventListener('click', function (e) {
        e.preventDefault();
        let data = new FormData();
        data.set('notes-for-solution', 'new submission');
        if (notesContent.value)
            data.set('existing-notes', notesContent.value);
        data.set('metadata', JSON.stringify({
            SolutionName: solutionName,
            EvaluatedMetrics1: results[0],
            EvaluatedMetrics2: results[1],
        }));
        fetch('', { method: 'post', body: data }).then(function (response) {
            return response.text();
        }).then(function (text) {
            let scrollOffset = window.scrollY + document.documentElement.clientHeight - scrollHeight();
            let dom = new DOMParser().parseFromString(text, 'text/html');
            let discardChangesButton = dom.getElementById('notes-discard-changes');
            let saveChangesButton = dom.getElementById('notes-save-changes');
            discardChangesButton.addEventListener('click', function (e) {
                e.preventDefault();
                document.getElementById('container').remove();
                document.getElementById('container-underneath').id = 'container';
                updateCountdowns();
                window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
            });
            saveChangesButton.addEventListener('click', function (e) {
                e.preventDefault();
                updateNotesButtonWithNotes(document.getElementById('edit-notes').value);
                document.getElementById('container').remove();
                document.getElementById('container-underneath').id = 'container';
                updateCountdowns();
                window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
            });
            let newContainer = dom.getElementById('container');
            let saveAndSubmit = dom.createElement('div');
            saveAndSubmit.innerHTML = `<input type='submit' value='save and submit' id='save-and-submit'>`;
            newContainer.appendChild(saveAndSubmit);
            dom.getElementById('save-and-submit').addEventListener('click', function (e) {
                e.preventDefault();
                notesContent.value = document.getElementById('edit-notes').value;
                beforeSubmit();
                document.getElementById('submission-form').submit();
            });
            document.getElementById('container').id = 'container-underneath';
            document.getElementById('container-underneath').after(newContainer);
            updateCountdowns();
        })
    });
    if (existingNotes[index])
        updateNotesButtonWithNotes(existingNotes[index]);
}
let submissionForm = document.getElementById('submission-form');
if (submissionForm) {
    let metrics = JSON.parse(submissionForm.dataset.metrics);
    let puzzleData;
    function htmlEscape(str) {
        return str.replace(/&/g, '&amp;').replace(/</g, '&lt;');
    }
    function updatePreviews() {
        if (!puzzleData)
            return;
        let scrollOffset = window.scrollY + document.documentElement.clientHeight - scrollHeight();
        let uploadBox = document.getElementById('submission-upload-box');
        let previews = document.querySelectorAll('.submission.preview');
        let existingNotes = [];
        for (let i = 0; i < previews.length; ++i) {
            if (previews[i].cancelEvaluation)
                previews[i].cancelEvaluation();
            let notes = submissionForm.elements[`notes-content-${i}`];
            if (notes)
                existingNotes[i] = notes.value;
            previews[i].parentNode.removeChild(previews[i]);
        }
        let solutionFiles = submissionForm.elements['solution'].files;
        for (let i = 0; i < solutionFiles.length; ++i) {
            let fileName = solutionFiles[i].name;
            let submission = document.createElement('div');
            submission.className = 'submission preview';
            submission.innerHTML = `
                <div class='info'>
                    <b class='submission-name'>${htmlEscape(fileName)}</b>
                    <div class='metrics'><b>now loading&hellip;</b></div>
                </div>`
            let metricsElement = submission.querySelector('.metrics');
            uploadBox.parentNode.insertBefore(submission, uploadBox);
            if (solutionFiles[i].size > 1000000) {
                metricsElement.className = 'error';
                metricsElement.textContent = 'solution files must be less than 1MB';
                continue;
            }
            let reader = new FileReader();
            reader.onload = function (e) {
                if (!submission.parentNode)
                    return;
                let scrollOffset = window.scrollY + document.documentElement.clientHeight - scrollHeight();
                let solutionData = new Uint8Array(reader.result);
                let solutionName = fileName;
                if (solutionData[0] === 7) {
                    let data = solutionData.slice(4);
                    function parseString() {
                        let n = 0;
                        let shift = 0;
                        while (data.length > 0) {
                            let a = data[0];
                            data = data.slice(1);
                            n |= (a & 0x7f) << shift;
                            shift += 7;
                            if ((a & 0x80) === 0)
                                break;
                        }
                        let str = new TextDecoder().decode(data.slice(0, n));
                        data = data.slice(n);
                        return str;
                    }
                    let puzzleFileName = parseString();
                    solutionName = parseString();
                    if (!('disablePuzzleFileNameValidation' in submissionForm.dataset) && puzzleFileName + '.puzzle' != document.getElementById('puzzle-link').download) {
                        metricsElement.className = 'error';
                        metricsElement.textContent = `this is a solution to a different puzzle, ${puzzleFileName}.puzzle`;
                        window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
                        return;
                    }
                    submission.querySelector('.submission-name').textContent = solutionName;
                }
                if (document.getElementById('showcase-checkbox').checked) {
                    submission.innerHTML = `
                        <div class='info'>
                            <b>${htmlEscape(solutionName)}</b> (showcase)
                        </div>
                        <div class='buttons'>
                            <button class='notes-button'>add notes</button>
                        </div>`;
                    renderNotesButton(submission, solutionName, [[],[]], existingNotes, i);
                } else {
                    let evaluatorMessageElement;
                    submission.cancelEvaluation = evaluateMetrics(puzzleData, solutionData, metrics, function (result) {
                        let scrollOffset = window.scrollY + document.documentElement.clientHeight - scrollHeight();
                        if (result.error) {
                            metricsElement.className = 'error';
                            metricsElement.textContent = result.error;
                        } else {
                            let metricHTML = '';
                            for (let i = 0; i < result.results.length; ++i) {
                                if (result.results[i].length === 0)
                                    continue;
                                if (metricHTML.length > 0)
                                    metricHTML += '<br>';
                                if (i < metrics.Metrics.length)
                                    metricHTML += `<b>${htmlEscape(metrics.Metrics[i].Name)}</b> `;
                                metricHTML += htmlEscape(result.results[i].map(formatResult).join('/'));
                            }
                            metricsElement.innerHTML = metricHTML;
                            let notesElement = document.createElement('div');
                            notesElement.className = 'buttons';
                            notesElement.innerHTML = `<button class='notes-button'>add notes</button>`;
                            submission.querySelector('.info').after(notesElement);
                            renderNotesButton(submission, solutionName, result.results, existingNotes, i);
                        }
                        window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
                    }, function (targetSelector, evaluatorMessage, attribute, value) {
                        let scrollOffset = null;
                        if (!evaluatorMessageElement) {
                            scrollOffset = window.scrollY + document.documentElement.clientHeight - scrollHeight();
                            evaluatorMessageElement = document.createElement('div');
                            evaluatorMessageElement.className = 'custom-evaluator-message';
                            metricsElement.parentNode.insertBefore(evaluatorMessageElement, metricsElement);
                        }
                        let target = evaluatorMessageElement;
                        if (targetSelector)
                            target = evaluatorMessageElement.querySelector(targetSelector);
                        if (typeof evaluatorMessage === 'string')
                            target.innerHTML = evaluatorMessage;
                        else if (typeof attribute === 'string')
                            target.setAttribute(attribute, value);
                        if (scrollOffset !== null)
                            window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
                    });
                }
                window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
            };
            reader.readAsArrayBuffer(solutionFiles[i]);
        }
        window.scrollTo(window.scrollX, scrollOffset - document.documentElement.clientHeight + scrollHeight());
    }
    fetch(document.getElementById('puzzle-link').href).then(function (response) {
        return response.arrayBuffer();
    }).then(function (data) {
        puzzleData = new Uint8Array(data);
    }).then(updatePreviews);
    submissionForm.addEventListener('submit', beforeSubmit);
    submissionForm.elements['solution'].addEventListener('click', function () {
        this.value = null;
        updatePreviews();
        window.onbeforeunload = function () {
            if (submissionForm.elements['solution'].files.length > 0)
                return 'there are unsubmitted solutions; are you sure you want to navigate away?';
        };
    });
    submissionForm.elements['solution'].addEventListener('change', updatePreviews);
    document.getElementById('showcase-checkbox').addEventListener('change', updatePreviews);
    function updateShowPrevious() {
        let submitter = submissionForm.elements['submitter'].value;
        let key = submissionForm.elements['secret-key'].value;
        submissionForm.elements['show-previous'].disabled = submitter.length == 0 || key.length < 8 || key.length > 300;
    }
    updateShowPrevious();
    submissionForm.elements['submitter'].addEventListener('input', updateShowPrevious);
    submissionForm.elements['secret-key'].addEventListener('input', updateShowPrevious);
    submissionForm.elements['show-previous'].addEventListener('click', function (e) {
        e.preventDefault();
        const hideTitle = 'hide previous submissions';
        if (submissionForm.elements['show-previous'].value === hideTitle) {
            document.getElementById('previous').innerHTML = '';
            submissionForm.elements['show-previous'].value = 'show previous submissions';
            return;
        }
        submissionForm.elements['show-previous'].value = hideTitle;
        let data = new FormData(submissionForm);
        data.set('show-previous', 'yes');
        fetch('', { method: 'post', body: data }).then(function (response) {
            return response.text();
        }).then(function (text) {
            if (submissionForm.elements['show-previous'].value !== hideTitle)
                return;
            let dom = new DOMParser().parseFromString(text, 'text/html');
            let newPrevious = dom.getElementById('previous');
            let oldPrevious = document.getElementById('previous');
            oldPrevious.parentNode.replaceChild(newPrevious, oldPrevious);
            localizeDatetimes();
            newPrevious.scrollIntoView(true);
        });
    });
}
// inspired by https://stackoverflow.com/a/49041392
function toSortKey(elem) {
    let text = elem.dataset.sort;
    if (!text)
        text = elem.textContent;
    if (!text) {
        let input = elem.querySelector('input');
        if (input) {
            if (input.checked)
                text = 'checked';
            else
                text = input.value;
        }
    }
    let key = [];
    while (text) {
        let whitespace = text.match(/^\s+/);
        if (whitespace)
            text = text.slice(whitespace[0].length);
        let number = text.match(/^\d+(\.\d+)?/);
        if (number) {
            key.push(parseFloat(number));
            text = text.slice(number[0].length);
        } else {
            key.push(text.slice(0, 1));
            text = text.slice(1);
        }
    }
    return key;
}
function compareSortKeys(a, b) {
    let aLoopStart = -1;
    let bLoopStart = -1;
    for (let i = 0, j = 0; i < a.length && j < b.length; ++i, ++j) {
        if (a[i] === '[' && i + 1 < a.length)
            aLoopStart = ++i;
        else if (aLoopStart >= 0 && a[i] === ']' && b[j] !== ']')
            i = aLoopStart;
        if (b[j] === '[' && j + 1 < b.length)
            bLoopStart = ++j;
        else if (bLoopStart >= 0 && b[j] === ']' && a[i] !== ']')
            j = bLoopStart;
        if (typeof a[i] === 'number' && typeof b[j] !== 'number')
            return -1;
        if (typeof a[i] !== 'number' && typeof b[j] === 'number')
            return 1;
        if (a[i] < b[j])
            return -1;
        if (a[i] > b[j])
            return 1;
    }
    return b.length - a.length;
}
let tables = document.querySelectorAll('table');
for (let i = 0; i < tables.length; ++i) {
    let tableHeaders = tables[i].querySelectorAll('th');
    let activeHeader;
    for (let j = 0; j < tableHeaders.length; ++j) {
        let direction = 1;
        tableHeaders[j].addEventListener('click', function (e) {
            e.preventDefault();
            if (activeHeader == j)
                direction = -direction;
            activeHeader = j;
            let rows = [...tables[i].querySelectorAll('tr')].slice(1);
            rows.sort(function (a, b) { return direction * compareSortKeys(toSortKey(a.children[j]), toSortKey(b.children[j])); });
            for (let k = 0; k < rows.length; ++k)
                rows[k].parentNode.appendChild(rows[k]);
        });
    }
}
let solutionFiles = document.querySelectorAll('a.solution-file');
let previewURL = 'https://ssk97.github.io/display/demo.html';
for (let i = 0; i < solutionFiles.length; ++i) {
    let previewLink = document.createElement('a');
    previewLink.textContent = 'preview';
    previewLink.href = previewURL;
    previewLink.target = '_blank';
    previewLink.addEventListener('click', function (e) {
        e.preventDefault();
        let w = window.open(previewURL);
        let solution;
        let puzzle;
        let ready = false;
        function loadPreview() {
            w.postMessage({ command: 'load', puzzle: puzzle, solution: solution}, '*');
        }
        window.addEventListener('message', function (e) {
            if (e.origin !== new URL(previewURL).origin)
                return;
            if (solution && puzzle)
                loadPreview();
            else
                ready = true;
        });
        fetch(solutionFiles[i].href).then(function (response) {
            return response.arrayBuffer();
        }).then(function (solutionData) {
            solution = new Uint8Array(solutionData);
            fetch(document.getElementById('table-form').dataset.puzzle).then(function (response) {
                return response.arrayBuffer();
            }).then(function (puzzleData) {
                puzzle = new Uint8Array(puzzleData);
                if (ready)
                    loadPreview();
            });
        })
    });
    solutionFiles[i].after(' (', previewLink, ')');
}
let passwords = document.querySelectorAll('input[type="password"]');
for (let i = 0; i < passwords.length; ++i) {
    let parent = passwords[i].parentNode;
    let container = document.createElement('div');
    container.className = 'password-container';
    container.innerHTML = `
        <svg height="30px" viewBox="0 0 283 263" fill="none" xmlns="http://www.w3.org/2000/svg">
        <title>show or hide secret key</title>
        <path d="M267.312 131C242.163 174.075 195.456 203 142 203C88.5438 203 41.8369 174.075 16.688 131C41.8369 87.9249 88.5438 59 142 59C195.456 59 242.163 87.9249 267.312 131Z" class="svg-stroke" stroke-width="10"/>
        <circle cx="142" cy="131" r="55" class="svg-stroke" stroke-width="10"/>
        <circle cx="142" cy="131" r="16" class="svg-fill"/>
        </svg>
    `;
    container.querySelector('svg').addEventListener('click', function () {
        console.log('a');
        if (container.classList.toggle('password-visible'))
            passwords[i].type = 'text';
        else
            passwords[i].type = 'password';
    });
    parent.replaceChild(container, passwords[i]);
    container.insertBefore(passwords[i], container.firstChild);
}
console.log('the source code for events.critelli.technology can be downloaded at <http://events.critelli.technology/static/events.zip>.  to view it in your web browser, go to <http://dezip.org/http://events.critelli.technology/static/events.zip>.');
