'use strict';

let libverify;
let heap;
let metricAddr;

let wasi = { proc_exit: function () { } };
if (typeof process !== 'undefined')
    wasi.proc_exit = process.exit;

function instantiate(wasmData) {
    return WebAssembly.instantiate(wasmData, {
        env: {
            emscripten_notify_memory_growth: function () {
                heap = new Uint8Array(libverify.memory.buffer);
            },
        },
        wasi_snapshot_preview1: wasi,
    }).then(function (result) {
        libverify = result.instance.exports;
        heap = new Uint8Array(libverify.memory.buffer);
        metricAddr = libverify.malloc(256);
    });
}

function getString(memory, address) {
    if (!address)
        return null;
    let end = address;
    while (memory[end] > 0)
        end++;
    return new TextDecoder().decode(memory.slice(address, end));
}

function getError(verifier) {
    return getString(heap, libverify.verifier_error(verifier.verifier));
}

function evaluateMetricExpression(verifier, expr) {
    expr = expr.toLowerCase();
    let precedence = { '+': 1, '-': 1, '*': 2, '/': 2, '**': 3, 'unary minus': 3 };
    let builtin = { 'min': Math.min, 'max': Math.max, 'sqrt': Math.sqrt, 'abs': Math.abs };
    let nickname = {
        'width': '"width*2"/2',
        'omniwidth': '"omniwidth*2"/2',
        'throughput': 'per repetition outputs / per repetition cycles',
        'rate': '1 / throughput',
        'gold': 'cost',
        'sum': 'cost + cycles + area',
        'sum4': 'sum + instructions',
        'arms': 'number of arms',
        'track': 'number of track segments',
        'latency': 'product 1 cycles',
        'default restrictions': 'overlap + max(0, parts of type baron - 1) + max(0, "parts of type glyph-disposal" - 1) + duplicate reagents + duplicate products + max(0, "maximum track gap^2" - 1)',
    };
    libverify.verifier_error_clear(verifier.verifier);
    let token = /\s*(?:(\(|\)|,|\+|-|\/|\(|\*\*?)|'([^']*)'|"([^"]*)"|((?:[^+*/(),'"\s-]+\s*)*[^+*/(),'"\s-]+))\s*/y;
    let values = [];
    let ops = [];
    function reduceWithPrecedence(prec) {
        while (ops.length > 0 && ops[ops.length-1] !== '(' && !(precedence[ops[ops.length-1]] < prec)) {
            var op = ops.pop();
            if (op instanceof Function)
                throw 'function is missing parentheses';
            let a;
            switch (op) {
                case '+': a = values.pop(); values.push(values.pop() + a); break;
                case '-': a = values.pop(); values.push(values.pop() - a); break;
                case '*': a = values.pop(); values.push(values.pop() * a); break;
                case '/': a = values.pop(); values.push(values.pop() / a); break;
                case '**': a = values.pop(); values.push(Math.pow(values.pop(), a)); break;
                case 'unary minus': values.push(-values.pop()); break;
                default: throw 'unknown operator ' + op;
            }
        }
    }
    let expectingTerm = true;
    while (true) {
        if (token.lastIndex === expr.length)
            break;
        let r = expr.match(token);
        if (r === null) {
            throw 'mismatched quotation marks';
            break;
        }
        switch (r[1]) {
        case undefined:
            if (!expectingTerm)
                throw 'expecting operator but found term';
            let m = r[2] || r[3] || r[4] || '';
            if (builtin[m]) {
                ops.push(builtin[m]);
                break;
            }
            let v;
            let match;
            if (nickname[m])
                v = evaluateMetricExpression(verifier, nickname[m]);
            else if (match = m.match(/^(minutes|seconds|milliseconds) elapsed since (\d{4}-\d{2}-\d{2}t\d{2}:\d{2}(?::\d{2}(?:\.\d{3})?)?(?:z|[\+|-]\d{2}:\d{2})?)$/)) {
                let ms = new Date() - new Date(match[2].toUpperCase());
                switch (match[1]) {
                case 'minutes':
                    v = Math.floor(ms / 1000 / 60);
                    break;
                case 'seconds':
                    v = Math.floor(ms / 1000);
                    break;
                case 'milliseconds':
                    v = Math.floor(ms);
                    break;
                }
            } else if (m === 'output intervals') {
                let n = libverify.verifier_number_of_output_intervals(verifier.verifier);
                let intervals = [];
                for (let i = 0; i < n; ++i)
                    intervals.push(libverify.verifier_output_interval(verifier.verifier, i));
                v = {
                    intervals: intervals,
                    repeatsAfter: libverify.verifier_output_intervals_repeat_after(verifier.verifier)
                };
            } else {
                v = Number(m);
                if (isNaN(v)) {
                    if (verifier.cache[m])
                        v = verifier.cache[m];
                    else {
                        let bytes = new TextEncoder().encode(m);
                        if (bytes.length > 255)
                            throw 'metric too long';
                        heap.set(bytes, metricAddr);
                        heap[metricAddr + bytes.length] = 0;
                        v = libverify.verifier_evaluate_approximate_metric(verifier.verifier, metricAddr);
                        let error = getError(verifier);
                        if (error)
                            throw 'while measuring ' + m + ': ' + error;
                        verifier.cache[m] = v;
                    }
                }
            }
            values.push(v);
            expectingTerm = false;
            break;
        case '(':
            if (!expectingTerm)
                throw 'expecting operator but found open parenthesis';
            if (ops[ops.length-1] instanceof Function)
                values.push([]);
            ops.push('(');
            break;
        case ')':
            if (expectingTerm)
                throw 'expecting term but found close parenthesis';
            reduceWithPrecedence(-1);
            if (ops.length === 0)
                throw 'mismatched parentheses';
            ops.pop();
            if (ops[ops.length-1] instanceof Function) {
                let v = values.pop();
                let a = values.pop();
                values.push(ops.pop()(...a, v));
            }
            break;
        case ',':
            if (expectingTerm)
                throw 'expecting term but found comma';
            reduceWithPrecedence(-1);
            if (ops.length === 0)
                throw 'comma outside of parentheses';
            let a = values.pop();
            if (!(values[values.length-1] instanceof Array))
                throw 'comma outside of function call';
            values[values.length-1].push(a);
            expectingTerm = true;
            break;
        case '-':
            if (expectingTerm) {
                ops.push('unary minus');
                break;
            }
            // fall through.
        default:
            if (expectingTerm)
                throw 'expecting term but found operator';
            reduceWithPrecedence(precedence[r[1]]);
            ops.push(r[1]);
            expectingTerm = true;
            break;
        }
    }
    if (expectingTerm)
        throw 'expecting term at end of expression';
    reduceWithPrecedence(-1);
    if (ops.length > 0)
        throw 'mismatched parentheses';
    return values.pop();
}

function evaluateMetrics(puzzleData, solutionData, metrics, postEvaluatorMessage) {
    for (let i = 0; i < metrics.Metrics.length; ++i)
        metrics.Metrics[i].Submetrics = metrics.Metrics[i].Submetrics.filter(function (m) {return m !== ''; });
    let puzzle = libverify.malloc(puzzleData.length);
    let puzzleLength = puzzleData.length;
    let solution = libverify.malloc(solutionData.length);
    let solutionLength = solutionData.length;
    heap.set(puzzleData, puzzle);
    heap.set(solutionData, solution);
    let verifier = {
        // avoid `verifier_create_from_bytes_without_copying` so that custom
        // evaluators can free puzzle/solution pointers without invalidating the
        // original verifier.
        verifier: libverify.verifier_create_from_bytes(puzzle, puzzleLength, solution, solutionLength),
        cache: {},
    };
    libverify.verifier_disable_limits(verifier.verifier);
    let results = [];
    let error = getError(verifier);
    let metricErrors = [];
    let htmlMessage = null;
    let usingCustomEvaluator = false;
    for (let i = 0; !error && i < metrics.Metrics.length; ++i) {
        let subresults = [];
        if (metrics.Metrics[i].CustomEvaluatorWASM) {
            usingCustomEvaluator = true;
            let bstr = atob(metrics.Metrics[i].CustomEvaluatorWASM);
            let buffer = new Uint8Array(bstr.length);
            for (let j = 0; j < buffer.length; ++j)
                buffer[j] = bstr.charCodeAt(j);
            let evaluatorHeap;
            let modifiedVerifier = null;
            let cycleLimit = 100000;
            function rebuildVerifier() {
                if (modifiedVerifier)
                    libverify.verifier_destroy(modifiedVerifier.verifier);
                modifiedVerifier = {
                    verifier: libverify.verifier_create_from_bytes_without_copying(puzzle, puzzleLength, solution, solutionLength),
                    cache: {},
                };
                libverify.verifier_set_cycle_limit(modifiedVerifier.verifier, cycleLimit);
            }
            let evaluator = new WebAssembly.Instance(new WebAssembly.Module(buffer), {
                env: {
                    getPuzzleData: function (address, maxLength) {
                        let data = puzzleData;
                        if (data.length > maxLength)
                            data = data.slice(0, maxLength);
                        evaluatorHeap.set(data, address);
                        return puzzleData.length;
                    },
                    setPuzzleData: function (address, length) {
                        libverify.free(puzzle);
                        puzzle = libverify.malloc(length);
                        puzzleLength = length;
                        heap.set(evaluatorHeap.slice(address, address + length), puzzle);
                        rebuildVerifier();
                    },
                    getSolutionData: function (address, maxLength) {
                        let data = solutionData;
                        if (data.length > maxLength)
                            data = data.slice(0, maxLength);
                        evaluatorHeap.set(data, address);
                        return solutionData.length;
                    },
                    setSolutionData: function (address, length) {
                        libverify.free(solution);
                        solution = libverify.malloc(length);
                        solutionLength = length;
                        heap.set(evaluatorHeap.slice(address, address + length), solution);
                        rebuildVerifier();
                    },
                    setCycleLimit: function (limit) {
                        cycleLimit = limit;
                        rebuildVerifier();
                    },
                    setFailsOnWrongOutput: function (whichOutput, fails) {
                        if (!modifiedVerifier)
                            return;
                        libverify.verifier_set_fails_on_wrong_output(modifiedVerifier.verifier, whichOutput, fails);
                    },
                    setFailsOnWrongOutputBonds: function (whichOutput, fails) {
                        if (!modifiedVerifier)
                            return;
                        libverify.verifier_set_fails_on_wrong_output_bonds(modifiedVerifier.verifier, whichOutput, fails);
                    },
                    wrongOutputIndex: function () {
                        if (!modifiedVerifier)
                            return -1;
                        return libverify.verifier_wrong_output_index(modifiedVerifier.verifier);
                    },
                    wrongOutputAtom: function (u, v) {
                        if (!modifiedVerifier)
                            return -1;
                        return libverify.verifier_wrong_output_atom(modifiedVerifier.verifier, u, v);
                    },
                    wrongOutputClear: function () {
                        if (!modifiedVerifier)
                            return;
                        libverify.verifier_wrong_output_clear(modifiedVerifier.verifier);
                    },
                    getNumberOfMetrics: function () {
                        return metrics.Metrics[i].Submetrics.length;
                    },
                    evaluateMetricForPuzzle: function (metricIndex) {
                        try {
                            return evaluateMetricExpression(modifiedVerifier || verifier, metrics.Metrics[i].Submetrics[metricIndex]);
                        } catch (e) {
                            metricErrors[i] = e.toString();
                            return -1;
                        }
                    },
                    evaluateMetricExpressionForPuzzle: function (expression) {
                        try {
                            return evaluateMetricExpression(modifiedVerifier || verifier, getString(evaluatorHeap, expression));
                        } catch (e) {
                            metricErrors[i] = e.toString();
                            return -1;
                        }
                    },
                    evaluateRestrictionsForPuzzle: function () {
                        try {
                            return evaluateMetricExpression(modifiedVerifier || verifier, metrics.Metrics[i].Restrictions);
                        } catch (e) {
                            metricErrors[i] = e.toString();
                            return -1;
                        }
                    },
                    setError: function (error) {
                        metricErrors[i] = getString(evaluatorHeap, error);
                    },
                    getError: function (address, maxLength) {
                        let encoded = metricErrors[i] ? new TextEncoder().encode(metricErrors[i]) : new Uint8Array();
                        let data = encoded;
                        if (data.length > maxLength - 1)
                            data = data.slice(0, maxLength - 1);
                        evaluatorHeap.set(data, address);
                        evaluatorHeap[address + data.length] = 0;
                        return encoded.length;
                    },
                    clearError: function () {
                        metricErrors[i] = null;
                    },
                    setMetric: function (metric, value) {
                        subresults[metric] = value;
                    },
                    supportsEvaluatorMessages: function () {
                        return !!postEvaluatorMessage;
                    },
                    setEvaluatorMessage: function (message) {
                        if (!postEvaluatorMessage)
                            return;
                        postEvaluatorMessage({ evaluatorMessage: getString(evaluatorHeap, message) });
                    },
                    modifyEvaluatorMessage: function (selector, message) {
                        if (!postEvaluatorMessage)
                            return;
                        postEvaluatorMessage({
                            targetSelector: getString(evaluatorHeap, selector),
                            evaluatorMessage: getString(evaluatorHeap, message),
                        });
                    },
                    modifyEvaluatorMessageAttribute: function (selector, attribute, value) {
                        if (!postEvaluatorMessage)
                            return;
                        postEvaluatorMessage({
                            targetSelector: getString(evaluatorHeap, selector),
                            evaluatorMessageAttribute: getString(evaluatorHeap, attribute),
                            evaluatorMessageValue: getString(evaluatorHeap, value),
                        });
                    },
                },
            });
            evaluatorHeap = new Uint8Array(evaluator.exports.memory.buffer);
            try {
                evaluator.exports.evaluate();
            } catch (e) {
                metricErrors[i] = e.toString();
            }
            if (modifiedVerifier)
                libverify.verifier_destroy(modifiedVerifier.verifier);
        } else {
            if (metrics.Metrics[i].Restrictions !== '') {
                try {
                    let r = evaluateMetricExpression(verifier, metrics.Metrics[i].Restrictions);
                    if (r !== 0)
                        metricErrors[i] = 'solution failed to conform to restrictions for ' + metrics.Metrics[i].Name + ' metric';
                } catch (e) {
                    metricErrors[i] = e.toString();
                }
            }
            for (let j = 0; !metricErrors[i] && j < metrics.Metrics[i].Submetrics.length; ++j) {
                if (metrics.Metrics[i].Submetrics[j] === '')
                    continue;
                try {
                    subresults.push(evaluateMetricExpression(verifier, metrics.Metrics[i].Submetrics[j]));
                } catch (e) {
                    metricErrors[i] = e.toString();
                }
            }
        }
        if (metricErrors[i])
            results.push([]);
        else
            results.push(subresults);
    }
    if (metricErrors[0] && (metricErrors[1] || results[1].length === 0))
        error = metricErrors[0];
    if (!error && !usingCustomEvaluator) {
        try {
            function validateParsedMetric(m, pm, name) {
                let r = evaluateMetricExpression(verifier, m);
                let pr = evaluateMetricExpression(verifier, pm);
                if (r !== pr)
                    throw 'mismatch in ' + m + ': the solution has ' + pr + ', but the simulation reported a value of ' + r;
            }
            if (!metrics.DisableSimulation) {
                validateParsedMetric('cycles', 'parsed cycles');
                validateParsedMetric('area', 'parsed area');
            }
            validateParsedMetric('cost', 'parsed cost');
            validateParsedMetric('instructions', 'parsed instructions');
        } catch (e) {
            error = e.toString();
        }
    }
    libverify.verifier_destroy(verifier.verifier);
    libverify.free(puzzle);
    libverify.free(solution);
    let r = { results: results, error: error, metricErrors: metricErrors };
    if (htmlMessage)
        r.htmlMessage = htmlMessage;
    return r;
}

if (typeof process !== 'undefined') {
    // we're running in node.
    const fs = require('fs');
    instantiate(fs.readFileSync('static/libverify.wasm')).then(advanceState);
    let state = 'reading-puzzle-length';
    let length;
    let puzzle;
    let metrics;
    let solution;
    function advanceState() {
        if (!libverify)
            return;
        while (true) {
            let bytes;
            switch (state) {
            case 'reading-puzzle-length':
                bytes = process.stdin.read(4);
                if (bytes === null)
                    return;
                length = bytes.readUInt32LE(0);
                state = 'reading-puzzle';
                break;
            case 'reading-puzzle':
                puzzle = process.stdin.read(length);
                if (puzzle === null)
                    return;
                state = 'reading-metrics-length';
                break;
            case 'reading-metrics-length':
                bytes = process.stdin.read(4);
                if (bytes === null)
                    return;
                length = bytes.readUInt32LE(0);
                state = 'reading-metrics';
                break;
            case 'reading-metrics':
                metrics = process.stdin.read(length);
                if (metrics === null)
                    return;
                metrics = JSON.parse(new TextDecoder().decode(metrics));
                state = 'reading-solution-length';
                break;
            case 'reading-solution-length':
                bytes = process.stdin.read(4);
                if (bytes === null)
                    return;
                length = bytes.readUInt32LE(0);
                state = 'reading-solution';
                break;
            case 'reading-solution':
                solution = process.stdin.read(length);
                if (solution === null)
                    return;
                process.stdout.write(JSON.stringify(evaluateMetrics(puzzle, solution, metrics, null)));
                state = 'reading-solution-length';
                break;
            }
        }
    }
    process.stdin.on('readable', advanceState);
    process.stdin.on('end', process.exit);
} else if (self.WorkerGlobalScope) {
    // we're running in a web worker.
    fetch('/static/libverify.wasm?' + Date.now()).then(function (response) {
        return response.arrayBuffer();
    }).then(instantiate).then(function () {
        for (let i = 0; i < queue.length; ++i)
            handleMessage(queue[i]);
        queue = [];
    });
    let queue = [];
    self.addEventListener('message', function (e) {
        if (libverify)
            handleMessage(e.data);
        else
            queue.push(e.data);
    });
    function handleMessage(m) {
        try {
            postMessage(evaluateMetrics(m.puzzle, m.solution, m.metrics, m.supportsEvaluatorMessages ? postMessage : null));
        } catch (e) {
            postMessage({ results: [], error: `javascript error while evaluating metrics: ${e.toString()}`, metricErrors: [] });
        }
    }
}
